from policyengine_us.model_api import *


class slcsp_family_tier_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "ACA family tier premium amount"
    unit = USD
    definition_period = MONTH
    defined_for = "slcsp_family_tier_applies"

    def formula(tax_unit, period, parameters):
        # Get the base premium
        base_cost = tax_unit.household("slcsp_age_0", period)

        # Get the family tier multiplier based on composition
        family_tier_multiplier = tax_unit(
            "slcsp_family_tier_multiplier", period
        )

        # Calculate the total premium amount for the family tier
        return base_cost * family_tier_multiplier



from policyengine_us.model_api import *


class mi_withheld_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Michigan withheld income tax"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        agi = person("adjusted_gross_income_person", period)
        p = parameters(period).gov.states.mi.tax.income
        # The MI standard deduction only applys for elderly
        # Wo do not apply deductions here
        return p.rate * agi



from policyengine_us.model_api import *


class PellGrantCalculationMethod(Enum):
    EFC = "efc"
    SAI = "sai"


class pell_grant_calculation_method(Variable):
    value_type = Enum
    possible_values = PellGrantCalculationMethod
    default_value = PellGrantCalculationMethod.SAI
    entity = TaxUnit
    definition_period = YEAR
    label = "Pell Grant calculation method"

    def formula(tax_unit, period, parameters):
        tax_units = tax_unit("tax_unit_id", period)
        return np.full_like(
            tax_units,
            PellGrantCalculationMethod.EFC,
            dtype=PellGrantCalculationMethod,
        )

    def formula_2024(tax_unit, period, parameters):
        tax_units = tax_unit("tax_unit_id", period)
        return np.full_like(
            tax_units,
            PellGrantCalculationMethod.SAI,
            dtype=PellGrantCalculationMethod,
        )



from policyengine_us.model_api import *


class snap_state_using_standard_utility_allowance(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    label = "Whether a state always uses the standard utility allowance"

    def formula(spm_unit, period, parameters):
        state = spm_unit.household("state_code", period)
        p = parameters(period).gov.usda.snap.income.deductions.utility

        return p.always_standard[state]



from policyengine_us.model_api import *


class co_charitable_contribution_subtraction_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Colorado charitable contribution subtraction"
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR0104AD_2022.pdf#page=1",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=12",
        "https://casetext.com/regulation/colorado-administrative-code/department-200-department-of-revenue/division-201-taxation-division/rule-1-ccr-201-2-income-tax/rule-39-22-1044m-charitable-contribution-subtraction-for-taxpayers-claiming-the-federal-standard-deduction",
        # C.R.S. 39-22-104(4)(m)(1)
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        # Only available to filers who do not itemize on their federal tax return.
        # The tax form instructions also limit to filers who do not deduct charitable deductions,
        # but this is a redundant criterion.
        return ~tax_unit("tax_unit_itemizes", period)



from policyengine_us.model_api import *


class tx_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Texas TANF resource test"
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1220-limits",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-2420-eligibility-requirements",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        countable_resources = spm_unit("tx_tanf_countable_resources", period)
        p = parameters(period).gov.states.tx.tanf.resources

        return countable_resources <= p.resource_limit



from policyengine_us.model_api import *


class enrolled_in_ebb(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Enrolled for Emergency Broadband Benefit"
    documentation = "Whether a SPM unit is already enrolled in the Emergency Broadband Benefit"
    definition_period = YEAR



from policyengine_us.model_api import *


class ma_ccfa_enrolled(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Whether the family is currently enrolled in Massachusetts Child Care Financial Assistance (CCFA)"
    definition_period = MONTH
    defined_for = StateCode.MA



from policyengine_us.model_api import *


class total_misc_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Total miscellaneous deductions subject to the AGI floor"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/67#b"

    adds = "gov.irs.deductions.itemized.misc.sources"



from policyengine_us.model_api import *


class uncapped_ssi(Variable):
    value_type = float
    entity = Person
    label = "Uncapped SSI"
    unit = USD
    documentation = "Maximum SSI, less countable income (can be below zero)."
    definition_period = YEAR
    defined_for = "is_ssi_eligible"

    def formula(person, period, parameters):
        amount = person("ssi_amount_if_eligible", period)
        countable_income = person("ssi_countable_income", period)
        return amount - countable_income



from policyengine_us.model_api import *


class mo_wftc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri Working Families Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://revisor.mo.gov/main/OneSection.aspx?section=143.177&bid=49978&hl="
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        rate = parameters(period).gov.states.mo.tax.income.credits.wftc.match
        return federal_eitc * rate



from policyengine_us.model_api import *


class ca_cdcc_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "CDCC credit rate replicated to include California limitations"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/about-ftb/data-reports-plans/Summary-of-Federal-Income-Tax-Changes/index.html#PL-117-2-9631"
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)

        year = period.start.year
        if year == 2021:
            period_adjusted = f"{year-1}-01-01"
        else:
            period_adjusted = f"{year}-01-01"

        p = parameters(period_adjusted).gov.irs.credits.cdcc

        # First phase-out
        excess_agi = max_(0, agi - p.phase_out.start)  # start
        increments = np.ceil(excess_agi / p.phase_out.increment)
        percentage_reduction = increments * p.phase_out.rate
        phased_out_rate = max_(
            p.phase_out.min, p.phase_out.max - percentage_reduction  # max
        )

        # Second phase-out
        second_excess_agi = max_(
            0, agi - p.phase_out.second_start
        )  # second_start
        second_increments = np.ceil(second_excess_agi / p.phase_out.increment)
        second_percentage_reduction = second_increments * p.phase_out.rate

        return max_(0, phased_out_rate - second_percentage_reduction)



from policyengine_us.model_api import *


class wic_category_str(Variable):
    value_type = str
    entity = Person
    label = "WIC category (string)"
    documentation = "WIC category variable, stored as a string"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("wic_category", period).decode_to_str()



from policyengine_us.model_api import *


class ny_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"
    defined_for = StateCode.NY

    adds = ["ny_itemized_deductions_max"]
    subtracts = ["ny_itemized_deductions_reduction"]



from policyengine_us.model_api import *


class nj_other_retirement_income_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey Other Retirement Income Exclusion"
    unit = USD
    documentation = "New Jersey other retirement income exclusion"
    definition_period = YEAR
    reference = (
        "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=21",
        "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-6-15/",
    )
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        # follows Worksheet D
        p = parameters(period).gov.states.nj.tax.income.exclusions.retirement

        # determine age eligibility for head and spouse
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        age_eligible = person("age", period) >= p.age_threshold
        eligible_person = age_eligible & head_or_spouse

        # calculate age-eligible pension income
        pension = person("nj_eligible_pension_income", period)
        pension_income = tax_unit.sum(eligible_person * pension)

        fraction = tax_unit("nj_retirement_exclusion_fraction", period)
        filing_status = tax_unit("filing_status", period)
        exclusion_cap = p.max_amount[filing_status]

        # calculate maximum exclusion
        total_income_person = age_eligible * person("nj_total_income", period)
        total_income = tax_unit.sum(total_income_person)
        maximum_exclusion = min_(fraction * total_income, exclusion_cap)

        # calculate unused pension exclusion
        used = min_(pension_income * fraction, exclusion_cap)
        unused_exclusion = max_(0, maximum_exclusion - used)

        # calculate earnings eligibility
        earnings = add(tax_unit, period, ["earned_income"])
        limit = p.other_retirement_income.earned_income_threshold
        earnings_eligible = earnings <= limit

        age_eligible_unit = tax_unit.any(eligible_person)
        # return unused exclusion if age and earnings eligible
        return age_eligible_unit * earnings_eligible * unused_exclusion



from policyengine_us.model_api import *


class is_optional_senior_or_disabled_asset_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Asset-eligibility for State’s optional Medicaid pathway for seniors or people with disabilities"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#m"

    def formula(person, period, parameters):
        #  Assets
        personal_assets = person("ssi_countable_resources", period)  # $
        tax_unit = person.tax_unit
        assets = tax_unit.sum(personal_assets)

        #  state info
        is_joint = tax_unit("tax_unit_is_joint", period)
        state = person.household("state_code_str", period)

        #  Parameters
        p = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.senior_or_disabled

        #  Asset limit
        asset_limit = where(
            is_joint,
            p.assets.limit.couple[state],
            p.assets.limit.individual[state],
        )

        return assets < asset_limit



from policyengine_us.model_api import *


class mi_exemptions_count(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan exemptions count"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",
        "https://www.legislature.mi.gov/Publications/TaxpayerGuide.pdf",
        "https://www.michigan.gov/-/media/Project/Websites/taxes/2022RM/IIT/MI-1040CR7.pdf",
        "https://www.michigan.gov/-/media/Project/Websites/taxes/2022RM/IIT/BOOK_MI-1040CR-7.pdf#page=7",
    )
    adds = [
        "tax_unit_size",
        "mi_disabled_exemption_eligible_person",
        "is_fully_disabled_service_connected_veteran",
    ]



from policyengine_us.model_api import *


class is_parent_for_medicaid_nfc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid parent non-financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        is_dependent = person("is_tax_unit_dependent", period)
        has_dependent_in_tax_unit = (
            person.tax_unit("tax_unit_count_dependents", period) > 0
        )
        return ~is_dependent & has_dependent_in_tax_unit



from policyengine_us.model_api import *


class tx_tanf_assistance_unit_size(Variable):
    value_type = int
    entity = SPMUnit
    label = "Texas TANF assistance unit size"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-104",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-220-tanf",
    )
    defined_for = StateCode.TX

    adds = [
        "tx_tanf_eligible_child",
        "tx_tanf_eligible_parent",
    ]



from policyengine_us.model_api import *


class dc_ccsp_attending_days_per_month(Variable):
    value_type = int
    entity = Person
    label = "DC Child Care Subsidy Program (CCSP) attending days per month"
    definition_period = MONTH
    defined_for = StateCode.DC



from policyengine_us.model_api import *


class is_runaway_child(Variable):
    value_type = bool
    entity = Person
    label = "Is runaway child"
    documentation = "Whether an individual under 18 years old leaves home or their legal residence without parental or guardian permission"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/34/11279#4"



from policyengine_us.model_api import *


class hi_salt_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii state and local tax deduction"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=18"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=32"  # total itemized deduction worksheet
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    # Hawaii's state income tax allows an itemized deduction for state income taxes paid.
    # Hawaii does not cap the SALT deduction as the US does.

    adds = ["state_and_local_sales_or_income_tax", "real_estate_taxes"]



from policyengine_us.model_api import *


class il_cta_free_ride_eligible(Variable):
    value_type = bool
    entity = Person
    label = (
        "Eligible for the Illinois Chicago Transit Authority Free Ride Program"
    )
    definition_period = YEAR
    defined_for = StateCode.IL
    reference = "https://www.transitchicago.com/reduced-fare-programs/#free"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.il.rta.cta.free_ride_program.age_threshold
        age = person("age", period)
        young_child = age <= p.young_child
        bap_eligible = person("il_bap_eligible", period)
        eligible_military_status = person(
            "il_cta_military_service_pass_eligible", period
        )
        return young_child | bap_eligible | eligible_military_status



from policyengine_us.model_api import *


class is_pregnant_for_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Pregnant people"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#l_1_A"

    formula = all_of_variables(
        ["is_pregnant_for_medicaid_fc", "is_pregnant_for_medicaid_nfc"]
    )



from policyengine_us.model_api import *


class me_sales_and_property_tax_fairness_credit_income(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Maine sales and property tax fairness credit total income"
    definition_period = YEAR
    reference = "https://legislature.maine.gov/statutes/36/title36sec5219-KK.html"  # 1. D.
    defined_for = StateCode.ME

    adds = "gov.states.me.tax.income.credits.fairness.income_sources"



from policyengine_us.model_api import *


class public_pension_income(Variable):
    value_type = float
    entity = Person
    label = "public pension income"
    unit = USD
    documentation = "Income from government employee pensions."
    definition_period = YEAR

    adds = [
        "tax_exempt_public_pension_income",
        "taxable_public_pension_income",
    ]



from policyengine_us.model_api import *


class county_str(Variable):
    value_type = str
    entity = Household
    label = "County (string)"
    documentation = "County variable, stored as a string"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("county", period).decode_to_str()



from policyengine_us.model_api import *


class hi_food_excise_credit_child_receiving_public_support(Variable):
    value_type = bool
    entity = Person
    label = "Child received support for the hawaii food excise credit"
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(person, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.credits.food_excise_tax
        child_support = person("child_support_received", period)
        # Test if the child receive more than half support from public agency
        # (tanf or social_security_survivor_benefits)
        government_payments = add(
            person, period, ["tanf_person", "social_security_survivors"]
        )
        total_income = government_payments + child_support
        # If zero total income, assign 0% public support.
        public_support_percent = np.zeros_like(total_income)
        mask = total_income > 0
        public_support_percent[mask] = (
            government_payments[mask] / total_income[mask]
        )
        return (
            public_support_percent > p.minor_child.support_proportion_threshold
        )



from policyengine_us.model_api import *


class ar_low_income_tax_joint(Variable):
    value_type = float
    entity = Person
    label = (
        "Arkansas low income tax when married couples are filing separately"
    )
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=29"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        agi = add(person.tax_unit, period, ["ar_agi_joint"])
        head = person("is_tax_unit_head", period)
        agi_attributed_to_head = agi * head
        p = parameters(
            period
        ).gov.states.ar.tax.income.rates.low_income_tax_tables
        filing_status_separate = person.tax_unit("filing_status", period)
        status = filing_status_separate.possible_values
        dependents = person.tax_unit("tax_unit_dependents", period)
        return select(
            [
                filing_status_separate == status.SINGLE,
                (filing_status_separate == status.HEAD_OF_HOUSEHOLD)
                & (dependents <= 1),
                (filing_status_separate == status.HEAD_OF_HOUSEHOLD)
                & (dependents > 1),
                filing_status_separate == status.SEPARATE,
                (filing_status_separate == status.SURVIVING_SPOUSE)
                & (dependents <= 1),
                (filing_status_separate == status.SURVIVING_SPOUSE)
                & (dependents > 1),
                (filing_status_separate == status.JOINT) & (dependents <= 1),
                (filing_status_separate == status.JOINT) & (dependents > 1),
            ],
            [
                p.single.calc(agi_attributed_to_head, right=True),
                p.head_of_household.no_or_one_dependent.calc(
                    agi_attributed_to_head, right=True
                ),
                p.head_of_household.two_or_more_dependents.calc(
                    agi_attributed_to_head, right=True
                ),
                # Separate filers are ineligible to use the low income tax tables
                np.inf,
                p.surviving_spouse.no_or_one_dependent.calc(
                    agi_attributed_to_head, right=True
                ),
                p.surviving_spouse.two_or_more_dependents.calc(
                    agi_attributed_to_head, right=True
                ),
                p.joint.no_or_one_dependent.calc(
                    agi_attributed_to_head, right=True
                ),
                p.joint.two_or_more_dependents.calc(
                    agi_attributed_to_head, right=True
                ),
            ],
            default=0,
        )



from policyengine_us.model_api import *


class nc_scca_parent_fee(Variable):
    value_type = int
    entity = SPMUnit
    label = (
        "North Carolina Subsidized Child Care Assistance Program parent fee"
    )
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.NC
    reference = "https://policies.ncdhhs.gov/wp-content/uploads/chapter-8-parental-fees-7.pdf#page=2"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.nc.ncdhhs.scca
        parent_fee_rate = p.parent_fee_rate.value

        family_monthly_income = spm_unit("nc_scca_countable_income", period)

        parent_fee = family_monthly_income * parent_fee_rate

        # Round the number and only keep the integer part
        return np.round(parent_fee).astype(int)



from policyengine_us.model_api import *


class sc_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina income tax before non-refundable credits"
    defined_for = StateCode.SC
    unit = USD
    definition_period = YEAR
    reference = "https://dor.sc.gov/forms-site/Forms/SC1040TT_2022.pdf"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("sc_taxable_income", period)
        p = parameters(period).gov.states.sc.tax.income.rates
        return p.calc(taxable_income)



from policyengine_us.model_api import *


class income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "income tax before credits"
    unit = USD
    documentation = "Total (regular + AMT) income tax liability before credits"

    adds = [
        "income_tax_main_rates",
        "capital_gains_tax",
        "alternative_minimum_tax",
    ]



from policyengine_us.model_api import *


class sc_dependent_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina dependent exemption"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/SC1040_2022.pdf#page=2",
        "https://www.scstatehouse.gov/code/t12c006.php",
        # SECTION 12-6-1140 (13)
    )
    defined_for = StateCode.SC

    def formula(tax_unit, period, parameters):
        # Get relevant parameter subtree. The amount for dependent exemption is the same amount as the the young_child's.
        p = parameters(period).gov.states.sc.tax.income.deductions.young_child
        # every dependent is eligible
        dependents = tax_unit("tax_unit_dependents", period)
        # Multiply by the amount per exemption.
        return dependents * p.amount



from policyengine_us.model_api import *


class substitution_elasticity(Variable):
    value_type = float
    entity = Person
    label = "substitution elasticity of labor supply"
    unit = "/1"
    definition_period = YEAR

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.simulation.labor_supply_responses.elasticities.substitution

        if p.all != 0:
            return p.all

        # TODO: Parametrise
        EARNINGS_DECILE_MARKERS = [
            0,
            14e3,
            28e3,
            39e3,
            50e3,
            61e3,
            76e3,
            97e3,
            138e3,
            1_726e3,
        ]

        raw_earnings = add(
            person,
            period,
            [
                "employment_income_before_lsr",
                "self_employment_income_before_lsr",
            ],
        )
        earnings = max_(raw_earnings, 0)
        earnings_decile = (
            np.searchsorted(EARNINGS_DECILE_MARKERS, earnings) + 1
        )

        tax_unit = person.tax_unit
        # Primary earner == highest earner in tax unit
        max_earnings_in_unit = tax_unit.max(earnings)
        is_primary_earner = earnings == max_earnings_in_unit

        elasticities = np.zeros_like(earnings)

        # Handle zero earnings first
        zero_earnings = earnings == 0
        elasticities[zero_earnings] = 0

        # For non-zero earnings, assign elasticities
        non_zero_earnings = earnings > 0

        if np.any(non_zero_earnings):
            # First assign primary earner elasticities by decile
            decile_elasticities = [
                p.by_position_and_decile.primary._children[str(i + 1)]
                for i in range(10)
            ]
            for i in range(10):
                mask = (
                    non_zero_earnings
                    & (earnings_decile == i + 1)
                    & is_primary_earner
                )
                elasticities[mask] = decile_elasticities[i]

            # Then assign secondary earner elasticity where applicable
            secondary_mask = non_zero_earnings & ~is_primary_earner
            elasticities[secondary_mask] = p.by_position_and_decile.secondary

        return elasticities



from policyengine_us.model_api import *


class il_tanf_initial_employment_deduction_person(Variable):
    value_type = float
    entity = Person
    label = "Illinois Temporary Assistance for Needy Families (TANF) initial employment deduction per person"
    unit = USD
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.141"
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.il.dhs.tanf.income.initial_employment_deduction

        gross_earned_income = person("il_tanf_gross_earned_income", period)
        fpg = person.spm_unit(
            "il_tanf_initial_employment_deduction_fpg", period
        )
        payment_level = person.spm_unit(
            "il_tanf_payment_level_for_initial_eligibility", period
        )
        initial_employment_deduction = max_(p.rate * fpg - payment_level, 0)
        is_employed = gross_earned_income > 0
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        ied_eligible_person = is_employed & is_head_or_spouse
        uncapped_ied = ied_eligible_person * initial_employment_deduction
        return min_(gross_earned_income, uncapped_ied)



from policyengine_us.model_api import *


class private_pension_income(Variable):
    value_type = float
    entity = Person
    label = "private pension income"
    unit = USD
    documentation = "Income from non-government employee pensions."
    definition_period = YEAR

    adds = [
        "tax_exempt_private_pension_income",
        "taxable_private_pension_income",
    ]



from policyengine_us.model_api import *


class snap_countable_earner(Variable):
    value_type = bool
    entity = Person
    label = "Countable income earner"
    documentation = "Whether this person's earned income is counted for SNAP"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/7/2014"

    def formula(person, period, parameters):
        return ~person("snap_excluded_child_earner", period)



from policyengine_us.model_api import *


class nc_scca(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina Subsidized Child Care Assistance Program"
    unit = USD
    definition_period = MONTH
    defined_for = "nc_scca_entry_eligible"

    def formula(spm_unit, period, parameters):
        parent_fee = spm_unit("nc_scca_parent_fee", period)
        childcare_expenses = spm_unit(
            "spm_unit_pre_subsidy_childcare_expenses", period
        )
        uncapped_amount = max_(childcare_expenses - parent_fee, 0)
        maximum_payment = spm_unit("nc_scca_maximum_payment", period)
        return min_(uncapped_amount, maximum_payment)



from policyengine_us.model_api import *


class nm_2021_income_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico 2021 income tax rebate"
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503708/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgDsPABwBKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        dependent_elsewhere = tax_unit("head_is_dependent_elsewhere", period)
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.nm.tax.income.rebates["2021_income"]
        income_eligible = agi < p.main.income_limit[filing_status]
        eligible = income_eligible & ~dependent_elsewhere
        amount_if_eligible = p.main.amount[filing_status]
        return eligible * amount_if_eligible



from policyengine_us.model_api import *


class person_tax_unit_id(Variable):
    value_type = int
    entity = Person
    label = "Unique reference for the tax unit of this person"
    definition_period = YEAR



from policyengine_us.model_api import *


class co_property_tax_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado property tax exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/colorado-revised-statutes/title-39-taxation/property-tax/exemptions/article-3-exemptions/part-2-property-tax-exemption-for-qualifying-seniors-and-disabled-veterans/section-39-3-203-property-tax-exemption-qualifications"
    defined_for = StateCode.CO



from policyengine_us.model_api import *


class investment_income_form_4952(Variable):
    value_type = float
    entity = TaxUnit
    label = "Investment income from Form 4952"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class bottled_gas_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "Bottled gas expense"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class ia_subtractions_consolidated(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa subtractions from taxable income for years on or after 2023"
    unit = USD
    definition_period = YEAR
    reference = "https://www.legis.iowa.gov/docs/code/422.7.pdf"
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ia.tax.income.taxable_income
        total_subtractions = add(tax_unit, period, p.subtractions)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class ct_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Connecticut property tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://portal.ct.gov/-/media/DRS/Forms/2021/Income/CT-1040-Online-Booklet_1221.pdf#page=30"
    defined_for = "ct_property_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ct_agi", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.ct.tax.income.credits.property_tax
        real_estate_taxes = add(tax_unit, period, ["real_estate_taxes"])
        max_credit = min_(real_estate_taxes, p.cap)
        excess = max_(agi - p.reduction.start[filing_status], 0)
        total_increments = np.ceil(
            excess / p.reduction.increment[filing_status]
        )
        reduction_percent = p.reduction.rate * total_increments
        reduction_amount = max_credit * reduction_percent
        return max_(max_credit - reduction_amount, 0)



from policyengine_us.model_api import *


class is_in_medicaid_medically_needy_category(Variable):
    value_type = bool
    entity = Person
    label = "In Medicaid medically needy category"
    documentation = "Whether this person is in a Medicaid category for which there is a medically needy pathway."
    definition_period = YEAR

    def formula(person, period, parameters):
        mn_categories = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.medically_needy.categories
        aged_threshold = parameters(
            period
        ).gov.ssa.ssi.eligibility.aged_threshold
        is_child = any_(person, period, mn_categories.child.child_categories)
        is_disabled = person("is_ssi_disabled", period)
        is_senior = person("age", period) >= aged_threshold
        is_pregnant = person("is_pregnant_for_medicaid_nfc", period)
        is_parent = person("is_parent_for_medicaid_nfc", period)
        state = person.household("state_code_str", period)
        mn = mn_categories
        categories = [
            mn.child.child,
            mn.disabled,
            mn.parent,
            mn.pregnant,
            mn.senior,
        ]
        category_covered = [parameter[state] > 0 for parameter in categories]
        in_category = [
            is_child,
            is_disabled,
            is_parent,
            is_pregnant,
            is_senior,
        ]
        return np.any(
            [
                person_in_category & category_is_covered
                for person_in_category, category_is_covered in zip(
                    in_category, category_covered
                )
            ]
        )



from policyengine_us.model_api import *


class nc_scca_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina Subsidized Child Care Assistance program countable income"
    unit = USD
    definition_period = MONTH
    reference = "https://policies.ncdhhs.gov/wp-content/uploads/FINAL-Chapter-7-Family-definition-and-determining-income-eligibility-08-05-24.pdf#page=11"
    defined_for = StateCode.NC
    adds = "gov.states.nc.ncdhhs.scca.income.sources"



from policyengine_us.model_api import *


class az_public_pension_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona Pension Exclusion"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2022_140BOOKLET.pdf#page=18"
        "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2021_140BOOKLET.pdf#page=23"
        "https://www.azleg.gov/ars/43/01022.htm"
    )
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.az.tax.income.subtractions.pension
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        pension_income = person("taxable_public_pension_income", period)
        eligible_pension_income = pension_income * head_or_spouse
        total_allowed_pension_exclusion = min_(
            p.public_pension_cap, eligible_pension_income
        )
        return tax_unit.sum(total_allowed_pension_exclusion)



from policyengine_us.model_api import *


class energy_efficient_insulation_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Energy efficient insulation expenditures"
    documentation = "Expenditures on any insulation material or system which is specifically and primarily designed to reduce the heat loss or gain of a dwelling unit when installed in or on such dwelling unit."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25C#c_3_A"



from policyengine_us.model_api import *


class ga_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia non-refundable dependent care credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        # Georgia matches the federal credit taken
        federal_cdcc = tax_unit("cdcc", period)
        rate = parameters(period).gov.states.ga.tax.income.credits.cdcc.rate
        return federal_cdcc * rate



from policyengine_us.model_api import *


class ar_post_secondary_education_tuition_deduction_person(Variable):
    value_type = float
    entity = Person
    label = "Arkansas person post-secondary education tuition deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/AR1075_2022.pdf#page=1"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ar.tax.income.deductions.itemized.tuition
        tuition_expenses = person("qualified_tuition_expenses", period)
        uncapped = p.rate * tuition_expenses
        cap = select(
            [
                person("technical_institution_student", period),
                person("four_year_college_student", period),
            ],
            [
                p.weighted_average_tuition.technical_institutes,
                p.weighted_average_tuition.four_year_college,
            ],
            default=p.weighted_average_tuition.two_year_college,
        )
        return min_(uncapped, cap)



from policyengine_us.model_api import *


class MassachusettsCCFACareProviderType(Enum):
    CENTER_BASED_CARE_EARLY_EDUCATION = "Center-Based Care Early Education"
    CENTER_BASED_CARE_SCHOOL_AGE = "Center-Based Care School Age"
    HEAD_START_PARTNER_AND_KINDERGARTEN = "Head Start Partner and Kindergarten"
    INFORMAL_CHILD_CARE = "Informal Child Care"
    FAMILY_CHILD_CARE = "Family Child Care"


class ma_ccfa_care_provider_type(Variable):
    value_type = Enum
    entity = Person
    possible_values = MassachusettsCCFACareProviderType
    default_value = (
        MassachusettsCCFACareProviderType.CENTER_BASED_CARE_SCHOOL_AGE
    )
    definition_period = MONTH
    defined_for = StateCode.MA
    label = "Massachusetts Child Care Financial Assistance (CCFA) care provider type"
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"



from policyengine_us.model_api import *


class ca_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "California deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ftb.ca.gov/forms/2021/2021-540.pdf"
        "https://www.ftb.ca.gov/forms/2022/2022-540.pdf"
    )
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        std_ded = tax_unit("ca_standard_deduction", period)
        itm_ded = tax_unit("ca_itemized_deductions", period)
        return where(itm_ded > std_ded, itm_ded, std_ded)



from policyengine_us.model_api import *


class ok_federal_eitc_demographic_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Meets demographic eligibility for EITC for the Oklahoma EITC computation"
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/oklahoma/title-68/section-68-2357-43/"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        has_child = tax_unit("eitc_child_count", period) > 0
        age = person("age", period)
        # Relative parameter reference break branching in some states that
        # modify EITC age limits.
        min_age_non_student = (
            parameters.gov.irs.credits.eitc.eligibility.age.min(f"2020-01-01")
        )
        min_age_student = (
            parameters.gov.irs.credits.eitc.eligibility.age.min_student(
                f"2020-01-01"
            )
        )
        student = person("is_full_time_student", period)
        min_age = where(student, min_age_student, min_age_non_student)
        max_age = parameters.gov.irs.credits.eitc.eligibility.age.max(
            f"2020-01-01"
        )
        meets_age_requirements = (age >= min_age) & (age <= max_age)
        return has_child | tax_unit.any(meets_age_requirements)



from policyengine_us.model_api import *


class ut_personal_exemption_additional_dependents(Variable):
    value_type = int
    entity = TaxUnit
    label = "Utah total additional dependents under the personal exemption"
    definition_period = YEAR
    reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1018.html"  # 59-10-1018 (1)(g)
    defined_for = StateCode.UT
    adds = ["ut_personal_exemption_additional_dependent_eligible"]



from policyengine_us.model_api import *


class id_income_tax_liable(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Liable to pay income taxes in Idaho"
    definition_period = YEAR
    reference = "https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_11-15-2021.pdf#page=10"
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        gross_income = add(tax_unit, period, ["irs_gross_income"])
        standard_deduction = tax_unit("standard_deduction", period)
        return gross_income > standard_deduction



from policyengine_us.model_api import *


class vt_military_retirement_pay_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Vermont military retirement income exclusion"
    reference = (
        "https://tax.vermont.gov/individuals/seniors-and-retirees",
        "https://legislature.vermont.gov/statutes/section/32/151/05830e",  # Legal Code Titl. 32 V.S.A. § 5830e (d)
    )
    unit = USD
    defined_for = StateCode.VT
    documentation = (
        "Vermont military retirement benefits exempt from Vermont taxation."
    )

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.vt.tax.income.agi.retirement_income_exemption
        # Get retirement amount from military retirement system
        tax_unit_military_retirement_pay = add(
            tax_unit, period, ["military_retirement_pay"]
        )

        # S.51 (2025): Income-based military pension exemption
        is_income_based_system = (
            p.military_retirement.income_based_structure.in_effect
        )
        agi = tax_unit("adjusted_gross_income", period)

        # Calculate exemption based on system type
        income_based_exemption = tax_unit(
            "vt_military_retirement_income_based_exemption", period
        )
        cap_based_exemption = tax_unit(
            "vt_military_retirement_cap_based_exemption", period
        )

        return where(
            is_income_based_system,
            income_based_exemption,
            cap_based_exemption,
        )



from policyengine_us.model_api import *


class ca_riv_general_relief_needs_standards(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Riverside County General Relief needs standards"
    definition_period = MONTH
    defined_for = "in_riv"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.ca.riv.general_relief.needs_standards
        size = spm_unit("ca_riv_general_relief_budget_unit_size", period)
        capped_size = clip(size, 1, 5)
        # Verify the housing cost using the DPSS 1140, GA Housing Information/Verification,
        # or any other sufficient verification, such as lease agreement
        housing_expense = spm_unit("housing_cost", period)
        capped_housing_benefit = min_(housing_expense, p.housing[capped_size])
        return (
            capped_housing_benefit
            + p.food[capped_size]
            + p.personal_needs[capped_size]
        )



from policyengine_us.model_api import *


class lifetime_learning_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Lifetime Learning Credit"
    unit = USD
    documentation = "Value of the non-refundable Lifetime Learning Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#c"

    def formula(tax_unit, period, parameters):
        credit_limit = tax_unit(
            "lifetime_learning_credit_credit_limit", period
        )
        potential = tax_unit("lifetime_learning_credit_potential", period)
        return min_(credit_limit, potential)



from policyengine_us.model_api import *


class ma_tafdc_child_support_deduction(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) child support deduction"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-250"  # GG
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        child_support = person("alimony_income", period)
        p = parameters(period).gov.states.ma.dta.tcap.deductions
        return min_(child_support, p.child_support_received)



from policyengine_us.model_api import *


class nc_tanf_reduced_need_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina TANF reduced need standard"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        income = add(
            spm_unit,
            period,
            [
                "nc_tanf_countable_earned_income",
                "nc_tanf_countable_gross_unearned_income",
            ],
        )
        need_standard = spm_unit("nc_tanf_need_standard", period)
        reduced_need_standard = max_(need_standard - income, 0)

        return reduced_need_standard



import pandas as pd
from pathlib import Path

FOLDER = Path(__file__).parent

medicaid_rating_areas = pd.read_csv(FOLDER / "medicaid_rating_areas.csv")
second_lowest_silver_plan_cost = pd.read_csv(
    FOLDER / "second_lowest_silver_plan_cost.csv"
)
aca_rating_areas = pd.read_csv(FOLDER / "aca_rating_areas.csv")



from policyengine_us.model_api import *


class pr_low_income_credit_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Puerto Rico low income credit"
    definition_period = YEAR
    reference = "https://casetext.com/statute/laws-of-puerto-rico/title-thirteen-taxation-and-finance/subtitle-17-internal-revenue-code-of-2011/part-ii-income-taxes/chapter-1007-credits-against-tax/subchapter-b-refundable-credits/30212-credit-for-low-income-individuals-older-than-sixty-five-65-years-of-age"
    defined_for = StateCode.PR

    def formula(person, period, parameters):
        age = person("age", period)
        p = parameters(period).gov.territories.pr.tax.income.credits.low_income
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        age_eligible = age >= p.age_threshold
        # if you claim earned income credit, aren't eligible for low income credit
        eitc_not_claimed = (
            person.tax_unit("pr_earned_income_credit", period) <= 0
        )
        return head_or_spouse & age_eligible & eitc_not_claimed



from policyengine_us.model_api import *


class ny_itemized_deductions_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY itemized deductions reduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"  # (f)&(g)
    defined_for = "ny_itemized_deductions_reduction_applies"

    def formula(tax_unit, period, parameters):
        reduction_based_on_charitable_deduction_applies = tax_unit(
            "ny_itemized_deductions_reduction_based_on_charitable_deduction_applies",
            period,
        )
        reduction_based_on_charitable_deduction = tax_unit(
            "ny_itemized_deductions_reduction_based_on_charitable_deduction",
            period,
        )
        incremental_reduction = tax_unit(
            "ny_itemized_deductions_incremental_reduction", period
        )
        return where(
            reduction_based_on_charitable_deduction_applies,
            reduction_based_on_charitable_deduction,
            incremental_reduction,
        )



from policyengine_us.model_api import *


class business_is_qualified(Variable):
    value_type = bool
    entity = Person
    label = "Business is qualified"
    unit = USD
    documentation = "Whether all income from self-employment, partnerships and S-corporations is from qualified businesses. A qualified trade or business is any trade or business other than a specified service trade or business, or employment. The list of specified service trades can be found at https://www.law.cornell.edu/uscode/text/26/1202#e_3_A."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#d_1"
    default_value = True



from policyengine_us.model_api import *


class ma_limited_income_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Limited Income Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/doc/2021-schedule-nts-l-nrpy-no-tax-status-and-limited-income-credit/download"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ma_agi", period)
        exemption_threshold = tax_unit(
            "ma_income_tax_exemption_threshold", period
        )
        income_over_threshold = max_(0, agi - exemption_threshold)
        # Compute AGI as a share of the exemption threshold.
        # Use a mask rather than where to avoid a divide-by-zero warning. Default to inf.
        income_ratio = np.ones_like(agi) * np.inf
        mask = exemption_threshold > 0
        income_ratio[mask] = agi[mask] / exemption_threshold[mask]
        p = parameters(period).gov.states.ma.tax.income.credits
        eligible = income_ratio <= p.limited_income_credit.income_limit
        tax_cap = p.limited_income_credit.percent * income_over_threshold
        income_tax = tax_unit("ma_income_tax_before_credits", period)
        excess_tax = max_(0, income_tax - tax_cap)
        return eligible * excess_tax



from policyengine_us.model_api import *


class co_tanf_countable_earned_income_need(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado TANF total countable income for need determination"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        gross_earnings = spm_unit(
            "co_tanf_countable_gross_earned_income", period
        )
        p = parameters(period).gov.states.co.cdhs.tanf.income.earned_exclusion
        enrolled = spm_unit("is_tanf_enrolled", period)
        annual_flat_exclusion = p.flat * MONTHS_IN_YEAR
        return where(
            enrolled,
            # For enrolled participants, Colorado applies only a percent exclusion.
            gross_earnings * (1 - p.percent),
            # For new enrollees, Colorado applies only a flat exclusion.
            max_(gross_earnings - annual_flat_exclusion, 0),
        )



from policyengine_us.model_api import *


class is_infant_for_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Infants"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#l_1_B"

    formula = all_of_variables(
        ["is_infant_for_medicaid_fc", "is_infant_for_medicaid_nfc"]
    )



from policyengine_us.model_api import *


class ssi_marital_earned_income(Variable):
    value_type = float
    entity = Person
    label = "Total SSI earned income for a marital unit"
    definition_period = YEAR

    def formula(person, period, parameters):
        both_eligible = person("ssi_marital_both_eligible", period)
        earned_income = person("ssi_earned_income", period)

        return where(
            both_eligible,
            person.marital_unit.sum(earned_income),
            earned_income,
        )



from policyengine_us.model_api import *


class ca_state_supplement_dependent_amount(Variable):
    value_type = float
    entity = SPMUnit
    label = "California SSI state supplement dependent amount"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ca.cdss.state_supplement.payment_standard
        person = spm_unit.members
        # Blind amount
        blind = person("is_blind", period)
        is_disabled = person("is_disabled", period)
        age = person("monthly_age", period)
        # Dependent amount
        dependent = person("is_tax_unit_dependent", period)
        dependent_age_eligible = age < p.dependent.age_limit
        dependent_count = spm_unit.sum(
            dependent & dependent_age_eligible & (is_disabled | blind)
        )
        return dependent_count * p.dependent.amount



from policyengine_us.model_api import *


class MassachusettsEAEDCLivingArrangement(Enum):
    A = "A"  # Living Arrangement A
    B = "B"  # Living Arrangement B
    C = "C"  # Living Arrangement C
    D = "D"  # Living Arrangement D
    E = "E"  # Living Arrangement E
    F = "F"  # Living Arrangement F
    H = "H"  # Living Arrangement H
    NONE = "None"


class ma_eaedc_living_arrangement(Variable):
    value_type = Enum
    entity = SPMUnit
    possible_values = MassachusettsEAEDCLivingArrangement
    default_value = MassachusettsEAEDCLivingArrangement.A
    definition_period = MONTH
    label = "Massachusetts EAEDC living arrangement"
    defined_for = StateCode.MA
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-435"
    )



from policyengine_us.model_api import *


class nc_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "North Carolina TANF eligible"
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        is_demographically_eligible = spm_unit(
            "nc_demographic_tanf_eligible", period
        )
        is_income_eligible = spm_unit("nc_tanf_income_eligible", period)

        return is_income_eligible & is_demographically_eligible



from policyengine_us.model_api import *


class ma_eaedc_countable_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts EAEDC countable earned income"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-500"  # (B) step 2

    def formula(spm_unit, period, parameters):
        total_earned_income_after_disregard = add(
            spm_unit, period, ["ma_eaedc_earned_income_after_disregard_person"]
        )
        # dependent care deduction
        dependent_care_deduction = add(
            spm_unit, period, ["ma_eaedc_dependent_care_deduction_person"]
        )

        return max_(
            total_earned_income_after_disregard - dependent_care_deduction, 0
        )



from policyengine_us.model_api import *


class il_use_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL use tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        # Tiered amount if income below a threshold, otherwise a percentage of AGI.
        agi = tax_unit("adjusted_gross_income", period)
        p = parameters(period).gov.states.il.tax.income.use_tax
        return p.amount.calc(agi) + (p.rate.calc(agi) * agi)



from policyengine_us.model_api import *


class nyc_school_tax_credit_fixed_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC School Tax Credit Fixed Amount"
    unit = USD
    definition_period = YEAR
    defined_for = "nyc_school_tax_credit_fixed_amount_eligible"

    def formula(tax_unit, period, parameters):
        # Get the tax unit's filing status.
        filing_status = tax_unit("filing_status", period)

        # Then get the School Tax Credit fixed amount part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.school.fixed

        return p.amount[filing_status]



from policyengine_us.model_api import *


class mt_refundable_credits(Variable):
    value_type = float
    entity = Person
    label = "Montana refundable credits"
    unit = USD
    reference = "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2022/12/Form-2-2022-Instructions.pdf#page=48"
    definition_period = YEAR
    defined_for = StateCode.MT
    adds = [
        "mt_refundable_credits_before_renter_credit",
        "mt_elderly_homeowner_or_renter_credit",
    ]
    # Under the gross income sources computation, the elderly homeowner or renter credit
    # is included in the list of refundable credits
    # This will create a potential circular reference (check reference Line7)
    # mt_refundable_credits_before_renter_credit was created to circumvent this



from policyengine_us.model_api import *


class student_loan_interest_ald_magi(Variable):
    value_type = float
    entity = Person
    label = "Modified adjusted gross income for the student loan interest ALD"
    unit = USD
    definition_period = YEAR
    documentation = "Above-the-line deduction for student loan interest"
    reference = "https://www.law.cornell.edu/uscode/text/26/221#b_2_C"
    defined_for = "student_loan_interest_ald_eligible"

    def formula(person, period, parameters):
        p_irs = parameters(period).gov.irs
        not_dependent = ~person("is_tax_unit_dependent", period)
        gross_income_sources = p_irs.gross_income.sources
        # Unemployment compensation is not considered in MAGI
        gross_income = [
            income_source
            for income_source in gross_income_sources
            if income_source
            not in p_irs.ald.student_loan_interest.magi.excluded_gross_income_sources
        ]
        total_gross_income = 0
        for source in gross_income:
            total_gross_income += not_dependent * max_(
                0, add(person, period, [source])
            )
        # Modified gross income is calculated with certain deductions excluded
        person_alds = p_irs.ald.student_loan_interest.magi.person_alds
        person_ald_vars = [f"{ald}_person" for ald in person_alds]
        ald_sum_person = add(person, period, person_ald_vars)
        all_alds = p_irs.ald.deductions
        other_alds = list(
            set(all_alds)
            - set(person_alds)
            - set(p_irs.ald.student_loan_interest.magi.excluded_alds)
        )
        ald_sum_taxunit = add(person.tax_unit, period, other_alds)
        filing_status = person.tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        frac = where(joint, 0.5, 1.0)
        ald_sum_taxunit_shared = not_dependent * ald_sum_taxunit * frac

        modified_adjusted_gross_income = (
            total_gross_income - ald_sum_person - ald_sum_taxunit_shared
        )
        if parameters(period).gov.contrib.ubi_center.basic_income.taxable:
            basic_income = person.tax_unit("basic_income", period)
            # split basic income evenly between head and spouse
            basic_income_shared = not_dependent * basic_income * frac
            modified_adjusted_gross_income += basic_income_shared
        # Combine the income for units filing jointly
        return modified_adjusted_gross_income



from policyengine_us.model_api import *


class md_qualifies_for_unmarried_childless_eitc(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Qualifies for the MD unmarried childless EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"  # (c)(3)
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        childless = tax_unit("eitc_child_count", period) == 0
        # Law says "individual".
        # Tax form instructions clarify that this means single/head/surviving spouse.
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        single_head_surviving_spouse = (
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE)
        )
        return childless & single_head_surviving_spouse



from policyengine_us.model_api import *


class de_pre_exclusions_agi(Variable):
    value_type = float
    entity = Person
    label = "Delaware individual adjusted gross income before exclusions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    adds = ["de_additions", "adjusted_gross_income_person"]
    subtracts = ["de_subtractions"]



from policyengine_us.model_api import *


class ok_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK
    adds = ["adjusted_gross_income", "ok_additions"]
    subtracts = ["ok_agi_subtractions"]



from policyengine_us.model_api import *


class medicaid_enrolled(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid enrolled"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a"
    defined_for = "is_medicaid_eligible"
    adds = ["takes_up_medicaid_if_eligible"]



from policyengine_us.model_api import *


class me_dependent_exemption_credit_amount_person(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Maine dependent exemption credit amount for each person"
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5219-SS.html"
    definition_period = YEAR
    defined_for = "ctc_qualifying_child"

    def formula(person, period, parameters):
        dependent = person("is_tax_unit_dependent", period)
        p = parameters(
            period
        ).gov.states.me.tax.income.credits.dependent_exemption
        age = person("age", period)
        multiplier = p.multiplier.calc(age)
        return dependent * multiplier * p.amount



from policyengine_us.model_api import *


class ny_itemized_deductions_higher_incremental_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York itemized deductions higher incremental reduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"  # (f)
    defined_for = "ny_itemized_deductions_reduction_applies"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ny.tax.income.deductions.itemized.reduction.incremental.higher
        agi = tax_unit("ny_agi", period)
        agi_excess = max_(agi - p.income_threshold, 0)
        numerator = min_(p.numerator, agi_excess)
        fraction = numerator / p.denominator
        itemized_deduction = tax_unit("ny_itemized_deductions_max", period)
        return itemized_deduction * p.rate * fraction



from policyengine_us.model_api import *


class ri_child_tax_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island Child Tax Rebate"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-08/H7123Aaa_CTR_0.pdf"
    defined_for = "ri_child_tax_rebate_eligible"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        age = person("age", period)
        dependent = person("is_tax_unit_dependent", period)
        p = parameters(
            period
        ).gov.states.ri.tax.income.credits.child_tax_rebate
        age_eligibility = age <= p.limit.age
        eligible_child = age_eligibility & dependent
        child_count = tax_unit.sum(eligible_child)
        capped_children = min_(child_count, p.limit.child)
        return capped_children * p.amount



from policyengine_us.model_api import *


class ny_qualified_geothermal_energy_system_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Qualified geothermal energy system equipment expenditures"
    documentation = "Money spent in the current year for the purchase or lease of geothermal energy system equipment."
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # PART 1 (g)(2)(C)(9)(g-4)
    defined_for = StateCode.NY



from policyengine_us.model_api import *


class tx_fpp_benefit(Variable):
    value_type = float
    entity = Person
    label = "Texas Family Planning Program benefit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.healthytexaswomen.org/healthcare-programs/family-planning-program/fpp-who-can-apply",
        "https://www.hhs.texas.gov/sites/default/files/documents/texas-womens-health-programs-report-2024.pdf",
    )
    defined_for = "tx_fpp_eligible"
    adds = ["gov.states.tx.fpp.annual_benefit"]



from policyengine_us.model_api import *


class vt_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont AGI additions"
    unit = USD
    documentation = "Additions to Vermont adjusted gross income"
    definition_period = YEAR
    defined_for = StateCode.VT
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112-2022.pdf#page=1"
        "https://legislature.vermont.gov/statutes/section/32/151/05811"
    )

    adds = ["tax_exempt_interest_income"]



from policyengine_us.model_api import *


class tx_dart_free_ride_eligible_young_child(Variable):
    value_type = bool
    entity = Person
    label = "Eligible young child for Dallas Area Rapid Transit (DART) Free Ride program"
    definition_period = YEAR
    defined_for = StateCode.TX
    reference = (
        "https://www.dart.org/fare/general-fares-and-overview/reduced-fares"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.states.tx.dart.reduced_fare.age_threshold
        age = person("age", period)
        # Under 5 years old - children under 5 ride free
        # The first threshold in the child parameter is 5 (the age when reduced fare starts)
        return age < p.child.thresholds[1]



from policyengine_us.model_api import *


class wv_homestead_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia homestead exemption"
    defined_for = StateCode.WV
    unit = USD
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-21/"

    def formula(tax_unit, period, parameters):
        assessed_property_value = add(
            tax_unit, period, ["assessed_property_value"]
        )
        p = parameters(
            period
        ).gov.states.wv.tax.income.exemptions.homestead_exemption

        return min_(assessed_property_value, p.max_amount)



from policyengine_us.model_api import *


class ne_child_care_subsidy_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Nebraska Child Care Subsidy program income eligible"
    definition_period = YEAR
    reference = (
        "https://nebraskalegislature.gov/laws/statutes.php?statute=68-1206",
        "https://dhhs.ne.gov/Pages/Child-Care-Parents.aspx",
    )
    defined_for = StateCode.NE

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ne.dhhs.child_care_subsidy.fpg_fraction
        income = add(spm_unit, period, ["adjusted_gross_income"])
        fpg = spm_unit("spm_unit_fpg", period)
        income_limit = fpg * p.initial_eligibility
        return income <= income_limit



from policyengine_us.model_api import *


class ar_misc_deduction_indiv(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas miscellaneous deduction when married filing separately"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/AR1075_2022.pdf#page=1"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.itemized.misc
        misc_ded = tax_unit("misc_deduction", period)
        agi = add(tax_unit, period, ["ar_agi_indiv"])
        return max_(
            0,
            misc_ded - p.floor * agi,
        )



from policyengine_us.model_api import *


class ms_pre_deductions_taxable_income_indiv(Variable):
    value_type = float
    entity = Person
    label = "Mississippi pre deductions taxable income when married couple file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=13",
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80105228.pdf",  # Line 38 - 49,
    )
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        agi = person("ms_agi", period)
        exemptions = person("ms_total_exemptions_indiv", period)
        return max_(agi - exemptions, 0)



from policyengine_us.model_api import *


class meets_ctc_identification_requirements(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Person meets CTC identification requirements"
    reference = (
        "https://www.congress.gov/bill/119th-congress/house-bill/1/text"
    )

    def formula(person, period, parameters):
        ssn_card_type = person("ssn_card_type", period)
        ssn_card_str = ssn_card_type.decode_to_str()
        p = parameters(period).gov.irs.credits.ctc
        return np.isin(ssn_card_str, p.eligible_ssn_card_type)



from policyengine_us.model_api import *


class reported_salt(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Reported State and local sales or income tax and real estate taxes subject to the SALT deduction limited to taxable income"
    unit = USD

    def formula(tax_unit, period, parameters):
        salt = tax_unit("salt", period)
        p = parameters(period).gov.simulation
        if p.limit_itemized_deductions_to_taxable_income:
            agi = tax_unit("adjusted_gross_income", period)
            exemptions = tax_unit("exemptions", period)
            agi_less_exemptions = max_(0, agi - exemptions)
            return min_(salt, agi_less_exemptions)
        return salt



from policyengine_us.model_api import *


class MassachusettsCCFARegion(Enum):
    WESTERN_CENTRAL_AND_SOUTHEAST = "Western, Central & Southeast"
    NORTHEAST = "Northeast"
    METRO_AND_BOSTON_METRO = "Metro & Boston Metro"


class ma_ccfa_region(Variable):
    value_type = Enum
    entity = Household
    possible_values = MassachusettsCCFARegion
    default_value = MassachusettsCCFARegion.WESTERN_CENTRAL_AND_SOUTHEAST
    definition_period = MONTH
    defined_for = StateCode.MA
    label = "Massachusetts Child Care Financial Assistance (CCFA) region"
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"

    def formula(household, period, parameters):
        county = household("county_str", period)

        p = parameters(period).gov.states.ma.eec.ccfa.region
        western_central_and_southeast = np.isin(
            county, p.western_central_and_southeast
        )
        northeast = np.isin(county, p.northeast)
        metro_and_boston_metro = np.isin(county, p.metro_and_boston_metro)

        conditions = [
            western_central_and_southeast,
            northeast,
            metro_and_boston_metro,
        ]
        results = [
            MassachusettsCCFARegion.WESTERN_CENTRAL_AND_SOUTHEAST,
            MassachusettsCCFARegion.NORTHEAST,
            MassachusettsCCFARegion.METRO_AND_BOSTON_METRO,
        ]

        return select(
            conditions,
            results,
            default=MassachusettsCCFARegion.WESTERN_CENTRAL_AND_SOUTHEAST,
        )



from policyengine_us.model_api import *


class taxsim_ssemp(Variable):
    value_type = float
    entity = TaxUnit
    label = "Self-employment income of spouse (excluding QBI)"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_primary = person("is_tax_unit_spouse", period)
        semp = person("self_employment_income", period)
        return tax_unit.sum(semp * is_primary)



from policyengine_us.model_api import *


class ny_drive_clean_rebate(Variable):
    value_type = float
    entity = Household
    label = "New York Drive Clean Rebate"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nyserda.ny.gov/-/media/Project/Nyserda/Files/Programs/Drive-Clean-NY/implementation-manual.pdf#page=8"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.nyserda.drive_clean
        # Calculate vehicle all-electric range
        vehicle_range = tax_unit(
            "ny_drive_clean_vehicle_electric_range", period
        )
        # Calculate vehicle price
        vehicle_cost = tax_unit("ny_drive_clean_vehicle_cost", period)
        # Calculate rebate size based on range
        rebate_amount_based_on_range = p.amount.calc(vehicle_range)
        # Determine whether or not vehicle is over the MSRP threhold,
        # qualifying for the flat rebate
        vehicle_over_msrp_threshold = (
            vehicle_cost >= p.flat_rebate.msrp_threshold
        )
        # If the vehicle is more expensive than the MSRP treshold, the filer
        # receives the flat rebate, otherwise they receive the range-based rebate
        return where(
            vehicle_over_msrp_threshold,
            p.flat_rebate.amount,
            rebate_amount_based_on_range,
        )



from policyengine_us.model_api import *


class capped_insulation_air_sealing_ventilation_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Capped insulation air sealing and ventilation rebate"
    documentation = "Before total high efficiency electric home rebate cap"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        percent_covered = tax_unit(
            "high_efficiency_electric_home_rebate_percent_covered", period
        )
        expenditures = add(
            tax_unit,
            period,
            # NB: Separated because insulation also receives a tax credit.
            [
                "energy_efficient_insulation_expenditures",
                "air_sealing_ventilation_expenditures",
            ],
        )
        cap = parameters(
            period
        ).gov.doe.high_efficiency_electric_home_rebate.cap.insulation_air_sealing_ventilation
        return min_(expenditures * percent_covered, cap)



from policyengine_us.model_api import *


class tx_dart_free_ride_benefit(Variable):
    value_type = float
    entity = Person
    label = "Dallas Area Rapid Transit (DART) free ride benefit value"
    unit = USD
    definition_period = YEAR
    defined_for = "tx_dart_free_ride_eligible_young_child"
    reference = (
        "https://www.dart.org/fare/general-fares-and-overview/reduced-fares"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.states.tx.dart.monthly_pass_cost
        # Currently only include the monthly pass costs
        return p.full_fare * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class ga_low_income_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia low income credit"
    unit = USD
    definition_period = YEAR
    reference = "https://dor.georgia.gov/document/document/2022-it-511-individual-income-tax-booklet/download"
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        # We follow the legal code, which says (in addition to head and spouse):
        # "multiplied by the number of dependents which the taxpayer is entitled to claim."
        # The tax form excludes adult dependents:
        # "Exemptions are self, spouse and natural or legally adopted children"
        exemptions = tax_unit("exemptions_count", period)
        p = parameters(period).gov.states.ga.tax.income.credits.low_income
        # age threshold
        age_threshold = p.supplement_age_eligibility
        aged_head = (tax_unit("age_head", period) >= age_threshold).astype(
            int
        )  # if so, return 1, otherwise return 0
        aged_spouse = (tax_unit("age_spouse", period) >= age_threshold).astype(
            int
        )  # if so, return 1, otherwise return 0
        aged_count = aged_head + aged_spouse
        total_exemptions = aged_count + exemptions
        federal_agi = tax_unit("adjusted_gross_income", period)
        amount_per_exemption = p.amount.calc(federal_agi)
        return total_exemptions * amount_per_exemption



from policyengine_us.model_api import *


class nj_taxable_income_before_property_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey taxable income before property tax deduction"
    unit = USD
    documentation = "NJ AGI less taxable income deductions, before property tax deduction (Line 39)"
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-3-1/"
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        agi = tax_unit("nj_agi", period)
        deductions = tax_unit("nj_total_deductions", period)
        exemptions = tax_unit("nj_total_exemptions", period)
        return max_(0, agi - deductions - exemptions)



from policyengine_us.model_api import *


class ms_agi(Variable):
    value_type = float
    entity = Person
    label = "Mississippi adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=14",
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80105228.pdf",  # Line 66
    )
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ms.tax.income
        gross_income = add(person, period, p.income_sources)
        adjustments = person("ms_agi_adjustments", period)
        net_income = max_(gross_income - adjustments, 0)
        # Allocate income from dependents to tax unit head.
        is_dependent = person("is_tax_unit_dependent", period)
        sum_dep_net_income = person.tax_unit.sum(is_dependent * net_income)
        is_head = person("is_tax_unit_head", period)
        return ~is_dependent * net_income + is_head * sum_dep_net_income



from .income_security_package import (
    income_security_package,
    create_income_security_package_reform,
)

# Keep individual reforms for backward compatibility but they all use the same implementation
baby_bonus_act = income_security_package
boost_act = income_security_package
end_child_poverty_act = income_security_package

__all__ = [
    "income_security_package",
    "create_income_security_package_reform",
    "baby_bonus_act",
    "boost_act",
    "end_child_poverty_act",
]



from policyengine_us.model_api import *


class co_income_qualified_senior_housing_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado Income Qualified Senior Housing Income Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/income-qualified-senior-housing-income-tax-credit",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=17",
    )
    defined_for = "co_income_qualified_senior_housing_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.co.tax.income.credits.income_qualified_senior_housing
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        max_amount = p.reduction.max_amount[filing_status]
        reduction_start = p.reduction.start
        increment = p.reduction.increment
        reduction_amount = p.reduction.amount[filing_status]
        excess = max_(agi - reduction_start, 0)
        increments = np.where(increment > 0, np.ceil(excess / increment), 0)
        total_reduction_amount = increments * reduction_amount
        return max_(max_amount - total_reduction_amount, 0)



from policyengine_us.model_api import *


class positive_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Positive Gross Income"
    unit = USD
    documentation = "Negative gross income values capped at zero"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        gross_income = add(tax_unit, period, ["irs_gross_income"])
        return max_(gross_income, 0)



from policyengine_us.model_api import *


class az_base_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona base standard deduction"
    unit = USD
    documentation = "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01041.htm"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.az.tax.income.deductions.standard
        filing_status = tax_unit("az_filing_status", period)
        return p.amount[filing_status]



from policyengine_us.model_api import *


class ne_child_care_subsidy_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Nebraska Child Care Subsidy program"
    definition_period = YEAR
    reference = (
        "https://nebraskalegislature.gov/laws/statutes.php?statute=68-1206",
        "https://dhhs.ne.gov/Pages/Child-Care-Subsidy-Information-for-Parents.aspx",
    )
    defined_for = StateCode.NE

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.ne.dhhs.child_care_subsidy
        person = spm_unit.members
        eligible_parent = person(
            "ne_child_care_subsidy_eligible_parent", period
        )
        eligible_parent_present = spm_unit.any(eligible_parent)
        eligible_child = person("ne_child_care_subsidy_eligible_child", period)
        eligible_child_present = spm_unit.any(eligible_child)
        income_eligible = spm_unit(
            "ne_child_care_subsidy_income_eligible", period
        )
        return (
            eligible_parent_present & eligible_child_present & income_eligible
        )



from policyengine_us.model_api import *


class ca_la_ez_save_countable_income(Variable):
    value_type = float
    entity = Household
    definition_period = YEAR
    label = "Los Angeles County EZ Save program countable income"
    defined_for = "in_la"

    adds = "gov.local.ca.la.dwp.ez_save.income_sources"



from policyengine_us.model_api import *


class meets_eitc_identification_requirements(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Person meets EITC identification requirements"
    reference = (
        # IRC 32(c)(1)(E) specifies filer and spouse needs a taxpayer identification number.
        "https://www.law.cornell.edu/uscode/text/26/32#c_1_E",
        # IRC 32(c)(3)(D)(i) specifies qualifying children need a taxpayer identification number.
        "https://www.law.cornell.edu/uscode/text/26/32#c_3_D_i",
        # IRC 32(m) defines a taxpayer identification number as a SSN,
        # except for those listed in sections II and III of SSA 205(c)(2)(B)(i).
        "https://www.law.cornell.edu/uscode/text/26/32#m",
        # SSA 205(c)(2)(B)(i) defines SSNs used for benefits.
        "https://www.ssa.gov/OP_Home/ssact/title02/0205.htm",
    )

    def formula(person, period, parameters):
        ssn_card_type = person("ssn_card_type", period)
        ssn_card_types = ssn_card_type.possible_values
        citizen = ssn_card_type == ssn_card_types.CITIZEN
        non_citizen_valid_ead = (
            ssn_card_type == ssn_card_types.NON_CITIZEN_VALID_EAD
        )
        return citizen | non_citizen_valid_ead



from policyengine_us.model_api import *


class tax_unit_ss_combined_income_excess(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxable Social Security combined income excess over base amount"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/86#b_1"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.irs.social_security.taxability.threshold.base

        # The legislation directs the usage of an income definition that is
        # a particularly modified AGI, plus half of gross Social Security
        # payments. Per IRC Section 86(b)(1), this fraction is always 0.5
        # and is handled by the combined_income_ss_fraction parameter.

        combined_income = tax_unit(
            "tax_unit_combined_income_for_social_security_taxability", period
        )
        filing_status = tax_unit("filing_status", period)

        status = filing_status.possible_values
        separate = filing_status == status.SEPARATE
        cohabitating = tax_unit("cohabitating_spouses", period)
        # Cohabitating married couples filing separately receive a base amount of 0
        base_amount = where(
            separate & cohabitating,
            p.separate_cohabitating,
            p.main[filing_status],
        )

        return max_(0, combined_income - base_amount)



from policyengine_us.model_api import *


class id_salt_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho SALT deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf#page=8"
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        state_and_local_tax = tax_unit(
            "state_and_local_sales_or_income_tax", period
        )
        # If the salt amount is above the federal cap, Idaho applies separate limitations
        filing_status = tax_unit("filing_status", period)
        cap = parameters(
            period
        ).gov.irs.deductions.itemized.salt_and_real_estate.cap[filing_status]
        # Federal Schedule A line 5b - State and local real estate taxes
        # Federal Schedule A line 5c - State and local personal property taxes
        real_estate_taxes = add(tax_unit, period, ["real_estate_taxes"])
        salt_and_real_estate_tax = state_and_local_tax + real_estate_taxes
        salt_above_cap = salt_and_real_estate_tax > cap
        reduced_cap = max_(cap - real_estate_taxes, 0)
        return where(salt_above_cap, reduced_cap, state_and_local_tax)



from policyengine_us.model_api import *


class va_age_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia age deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=16"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.va.tax.income.subtractions.age_deduction

        filing_status = tax_unit("filing_status", period)

        age_head = tax_unit("age_head", period)
        age_spouse = tax_unit("age_spouse", period)
        birth_year_head = period.start.year - age_head
        birth_year_spouse = period.start.year - age_spouse

        agi = tax_unit("va_age_deduction_agi", period)

        # Calculate the number of people who are eligible for an age deduction
        head_eligible = age_head >= p.age_minimum
        spouse_eligible = age_spouse >= p.age_minimum
        count_eligible = head_eligible.astype(int) + spouse_eligible.astype(
            int
        )

        # Calculate the number of people eligible for a full age deduction
        head_eligible_for_full_deduction = (
            birth_year_head < p.birth_year_limit_for_full_amount
        )
        spouse_eligible_for_full_deduction = (
            birth_year_spouse < p.birth_year_limit_for_full_amount
        )
        count_eligible_for_full_deduction = (
            head_eligible_for_full_deduction.astype(int)
            + spouse_eligible_for_full_deduction.astype(int)
        )

        # Calculate the maximum allowable deduction amount per filing
        maximum_allowable_deduction = p.amount * count_eligible

        # Calculate the amount that the adjusted federal AGI exceeds the threshold
        excess = max_(agi - p.threshold[filing_status], 0)

        # Reduce by the entire excess, unless both head and spouse (or head only if single)
        # are eligible for the full deduction
        reduction = excess * where(
            count_eligible == count_eligible_for_full_deduction, 0, 1
        )

        return max_(maximum_allowable_deduction - reduction, 0)



from policyengine_us.model_api import *


class amt_non_agi_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Income considered for AMT but not AGI"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class ma_covid_19_essential_employee_premium_pay_program(Variable):
    value_type = float
    entity = Person
    label = "MA COVID 19 Essential Employee Premium Pay Program"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/covid-19-essential-employee-premium-pay-program"
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        earnings = person("earned_income", period)
        tax_unit = person.tax_unit
        agi = tax_unit("adjusted_gross_income", period)
        fpg = tax_unit("tax_unit_fpg", period)
        poverty_ratio = agi / fpg
        p = parameters(
            period
        ).gov.states.ma.tax.income.credits.covid_19_essential_employee_premium_pay_program
        meets_earnings_test = earnings >= p.min_earnings
        meets_poverty_ratio_test = poverty_ratio <= p.max_poverty_ratio
        meets_program_test = add(person, period, p.disqualifying_programs) == 0
        eligible = (
            meets_earnings_test & meets_poverty_ratio_test & meets_program_test
        )
        return eligible * p.amount



from policyengine_us.model_api import *


class ar_additional_tax_credit_for_qualified_individuals(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas additional tax credit for qualified individuals"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR

    adds = ["ar_additional_tax_credit_for_qualified_individuals_person"]



from policyengine_us.model_api import *


class ri_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island AGI"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022%201041%20Schedule%20M_w.pdf"
    defined_for = StateCode.RI
    adds = ["adjusted_gross_income", "ri_additions"]
    subtracts = ["ri_subtractions"]



from policyengine_us.model_api import *


class mn_charity_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota charity subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.taxformfinder.org/forms/2021/2021-minnesota-form-m1m.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1m_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        charity = tax_unit("charitable_deduction", period)
        itemizing = tax_unit("mn_itemizing", period)
        p = parameters(period).gov.states.mn.tax.income.subtractions.charity
        subtraction_amount = p.fraction * max_(0, charity - p.threshold)
        return where(itemizing, 0, subtraction_amount)



from policyengine_us.model_api import *


def create_ny_working_families_tax_credit() -> Reform:
    class ny_working_families_tax_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "New York Working Families Tax Credit"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.NY

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ny.wftc
            income = tax_unit("ny_agi", period)
            filing_status = tax_unit("filing_status", period)
            joint = filing_status == filing_status.possible_values.JOINT
            reduction = where(
                joint,
                p.reduction.married.calc(income),
                p.reduction.single.calc(income),
            )
            children = tax_unit("ny_wftc_eligible_children", period)
            max_amount = p.amount.max * children
            min_amount = p.amount.min * children
            return max_(min_amount, max_amount - reduction)

    class ny_wftc_eligible_child(Variable):
        value_type = bool
        entity = Person
        label = "New York Working Families Tax Credit"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.NY

        def formula(person, period, parameters):
            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)
            p = parameters(period).gov.contrib.states.ny.wftc
            age_eligible = age <= p.child_age_threshold
            return is_dependent & age_eligible

    class ny_wftc_eligible_children(Variable):
        value_type = float
        entity = TaxUnit
        label = "New York Working Families Tax Credit"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.NY

        adds = ["ny_wftc_eligible_child"]

    # Replicating the relevant parts of the EITC for for younger
    # and older dependents separately

    class is_younger_child_dependent(Variable):
        value_type = bool
        entity = Person
        label = "Is a younger child dependents"
        reference = "https://legislation.nysenate.gov/pdf/bills/2023/S277C"
        definition_period = YEAR

        def formula(person, period, parameters):
            is_dependent = person("is_tax_unit_dependent", period)
            age = person("age", period)
            p = parameters(period).gov.contrib.states.ny.wftc
            age_eligible = age <= p.child_age_threshold
            return is_dependent & age_eligible

    class eitc_younger_children_count(Variable):
        value_type = int
        entity = TaxUnit
        label = "EITC-qualifying younger children"
        unit = USD
        documentation = "Number of children qualifying as children for the EITC, excluding dependents over 18."
        definition_period = YEAR

        adds = ["is_younger_child_dependent"]

    class eitc_younger_maximum(Variable):
        value_type = float
        entity = TaxUnit
        label = "Maximum EITC for younger children"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
        unit = USD

        def formula(tax_unit, period, parameters):
            child_count = tax_unit("eitc_younger_children_count", period)
            eitc = parameters(period).gov.irs.credits.eitc
            # We will reduce the maximum credit amount by the amount for self
            # as it is attributed to the younger children EITC
            # We also need to reduce it by the amount attributed to the younger children
            return eitc.max.calc(child_count)

    class eitc_younger_phase_in_rate(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-in rate"
        unit = "/1"
        documentation = "Rate at which the EITC phases in with income."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            child_count = tax_unit("eitc_younger_children_count", period)
            eitc = parameters(period).gov.irs.credits.eitc
            return eitc.phase_in_rate.calc(child_count)

    class eitc_younger_phased_in(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-in amount"
        unit = USD
        documentation = "EITC maximum amount, taking into account earnings."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            maximum = tax_unit("eitc_younger_maximum", period)
            phase_in_rate = tax_unit("eitc_younger_phase_in_rate", period)
            earnings = tax_unit("filer_adjusted_earnings", period)
            phased_in_amount = earnings * phase_in_rate
            return min_(maximum, phased_in_amount)

    class eitc_younger_phase_out_start(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-out start"
        unit = USD
        documentation = "Earnings above this level reduce EITC entitlement."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            count_children = tax_unit("eitc_younger_children_count", period)
            eitc = parameters(period).gov.irs.credits.eitc
            is_joint = tax_unit("tax_unit_is_joint", period)
            joint_bonus = eitc.phase_out.joint_bonus.calc(count_children)
            phase_out_start = eitc.phase_out.start.calc(count_children)
            return phase_out_start + is_joint * joint_bonus

    class eitc_younger_phase_out_rate(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-out rate"
        unit = USD
        documentation = "Percentage of earnings above the phase-out threshold that reduce the EITC."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            eitc = parameters(period).gov.irs.credits.eitc
            num_children = tax_unit("eitc_younger_children_count", period)
            return eitc.phase_out.rate.calc(num_children)

    class eitc_younger_reduction(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC reduction"
        unit = USD
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a_2"

        def formula(tax_unit, period, parameters):
            earnings = tax_unit("filer_adjusted_earnings", period)
            agi = tax_unit("adjusted_gross_income", period)
            highest_income_variable = max_(earnings, agi)
            phase_out_start = tax_unit("eitc_younger_phase_out_start", period)
            phase_out_rate = tax_unit("eitc_younger_phase_out_rate", period)
            phase_out_region = max_(
                0, highest_income_variable - phase_out_start
            )
            return phase_out_rate * phase_out_region

    class eitc_younger_demographic_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Meets demographic eligibility for EITC"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#c_1_A"

        def formula(tax_unit, period, parameters):
            return tax_unit("eitc_younger_children_count", period) > 0

    class younger_eitc_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Eligible for EITC"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#c_1_A"

        def formula(tax_unit, period, parameters):
            eitc = parameters.gov.irs.credits.eitc(period)
            investment_income_eligible = tax_unit(
                "eitc_investment_income_eligible", period
            )
            demographic_eligible = tax_unit(
                "eitc_younger_demographic_eligible", period
            )
            no_older_children = (
                tax_unit("eitc_older_children_count", period) == 0
            )
            # Define eligibility before considering separate filer limitation.
            eligible = (
                demographic_eligible
                & investment_income_eligible
                & no_older_children
            )
            # This parameter is true if separate filers are eligible.
            if eitc.eligibility.separate_filer:
                return eligible
            # If separate filers are not eligible, check if the filer is separate.
            filing_status = tax_unit("filing_status", period)
            separate = filing_status == filing_status.possible_values.SEPARATE
            return eligible & ~separate

    class younger_eitc(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Federal earned income credit"
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
        unit = USD
        defined_for = "younger_eitc_eligible"

        def formula(tax_unit, period, parameters):
            maximum = tax_unit("eitc_younger_maximum", period)
            phased_in = tax_unit("eitc_younger_phased_in", period)
            reduction = tax_unit("eitc_younger_reduction", period)
            limitation = max_(0, maximum - reduction)
            return min_(phased_in, limitation)

    class is_older_child_dependent(Variable):
        value_type = bool
        entity = Person
        label = "Is a older child dependents"
        reference = "https://legislation.nysenate.gov/pdf/bills/2023/S277C"
        definition_period = YEAR
        defined_for = StateCode.NY

        def formula(person, period, parameters):
            is_dependent = person("is_tax_unit_dependent", period)
            age = person("age", period)
            p_irs = parameters(period).gov.irs.dependent.ineligible_age
            p_ref = parameters(period).gov.contrib.states.ny.wftc
            wftc_older_age_eiligble = age > p_ref.child_age_threshold
            non_student_age_eligible = wftc_older_age_eiligble & (
                age < p_irs.non_student
            )
            student = person("is_full_time_student", period)
            student_age_eligible = (
                (age < p_irs.student) & student & wftc_older_age_eiligble
            )
            age_eligible = non_student_age_eligible | student_age_eligible
            return is_dependent & age_eligible

    class eitc_older_children_count(Variable):
        value_type = int
        entity = TaxUnit
        label = "EITC-qualifying younger children"
        documentation = "Number of children qualifying as children for the EITC, excluding dependents over 18."
        definition_period = YEAR

        adds = ["is_older_child_dependent"]

    class eitc_older_maximum(Variable):
        value_type = float
        entity = TaxUnit
        label = "Maximum EITC for younger children"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
        unit = USD

        def formula(tax_unit, period, parameters):
            older_child_count = tax_unit("eitc_older_children_count", period)
            younger_child_count = tax_unit(
                "eitc_younger_children_count", period
            )
            child_count = younger_child_count + older_child_count
            eitc = parameters(period).gov.irs.credits.eitc
            # We will reduce the maximum credit amount by the amount for self
            # as it is attributed to the younger children EITC
            # We also need to reduce it by the amount attributed to the younger children
            return eitc.max.calc(child_count)

    class eitc_older_phase_in_rate(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-in rate"
        unit = "/1"
        documentation = "Rate at which the EITC phases in with income."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            older_child_count = tax_unit("eitc_older_children_count", period)
            younger_child_count = tax_unit(
                "eitc_younger_children_count", period
            )
            child_count = younger_child_count + older_child_count
            eitc = parameters(period).gov.irs.credits.eitc
            return eitc.phase_in_rate.calc(child_count)

    class eitc_older_phased_in(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-in amount"
        unit = USD
        documentation = "EITC maximum amount, taking into account earnings."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            maximum = tax_unit("eitc_older_maximum", period)
            phase_in_rate = tax_unit("eitc_older_phase_in_rate", period)
            earnings = tax_unit("filer_adjusted_earnings", period)
            phased_in_amount = earnings * phase_in_rate
            return min_(maximum, phased_in_amount)

    class eitc_older_phase_out_start(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-out start"
        unit = USD
        documentation = "Earnings above this level reduce EITC entitlement."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            older_child_count = tax_unit("eitc_older_children_count", period)
            younger_child_count = tax_unit(
                "eitc_younger_children_count", period
            )
            count_children = younger_child_count + older_child_count
            eitc = parameters(period).gov.irs.credits.eitc
            is_joint = tax_unit("tax_unit_is_joint", period)
            joint_bonus = eitc.phase_out.joint_bonus.calc(count_children)
            phase_out_start = eitc.phase_out.start.calc(count_children)
            return phase_out_start + is_joint * joint_bonus

    class eitc_older_phase_out_rate(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-out rate"
        unit = USD
        documentation = "Percentage of earnings above the phase-out threshold that reduce the EITC."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            eitc = parameters(period).gov.irs.credits.eitc
            older_child_count = tax_unit("eitc_older_children_count", period)
            younger_child_count = tax_unit(
                "eitc_younger_children_count", period
            )
            count_children = younger_child_count + older_child_count
            return eitc.phase_out.rate.calc(count_children)

    class eitc_older_reduction(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC reduction"
        unit = USD
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a_2"

        def formula(tax_unit, period, parameters):
            earnings = tax_unit("filer_adjusted_earnings", period)
            agi = tax_unit("adjusted_gross_income", period)
            highest_income_variable = max_(earnings, agi)
            phase_out_start = tax_unit("eitc_older_phase_out_start", period)
            phase_out_rate = tax_unit("eitc_older_phase_out_rate", period)
            phase_out_region = max_(
                0, highest_income_variable - phase_out_start
            )
            return phase_out_rate * phase_out_region

    class eitc_older_demographic_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Meets demographic eligibility for EITC"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#c_1_A"

        def formula(tax_unit, period, parameters):
            person = tax_unit.members
            has_child = tax_unit("eitc_older_children_count", period) > 0
            age = person("age", period)
            # Relative parameter reference break branching in some states that
            # modify EITC age limits.
            min_age_student = (
                parameters.gov.irs.credits.eitc.eligibility.age.min_student(
                    period
                )
            )
            max_age = parameters.gov.irs.credits.eitc.eligibility.age.max(
                period
            )
            meets_age_requirements = (age >= min_age_student) & (
                age <= max_age
            )
            return has_child | tax_unit.any(meets_age_requirements)

    class older_eitc_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Eligible for EITC"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#c_1_A"

        def formula(tax_unit, period, parameters):
            eitc = parameters.gov.irs.credits.eitc(period)
            investment_income_eligible = tax_unit(
                "eitc_investment_income_eligible", period
            )
            demographic_eligible = tax_unit(
                "eitc_older_demographic_eligible", period
            )
            younger_eitc_eligible = tax_unit("younger_eitc_eligible", period)
            # Define eligibility before considering separate filer limitation.
            eligible = (
                demographic_eligible
                & investment_income_eligible
                & ~younger_eitc_eligible
            )
            # This parameter is true if separate filers are eligible.
            if eitc.eligibility.separate_filer:
                return eligible
            # If separate filers are not eligible, check if the filer is separate.
            filing_status = tax_unit("filing_status", period)
            separate = filing_status == filing_status.possible_values.SEPARATE
            return eligible & ~separate

    class older_eitc(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Federal earned income credit"
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
        unit = USD
        defined_for = "older_eitc_eligible"

        def formula(tax_unit, period, parameters):
            maximum = tax_unit("eitc_older_maximum", period)
            phased_in = tax_unit("eitc_older_phased_in", period)
            reduction = tax_unit("eitc_older_reduction", period)
            limitation = max_(0, maximum - reduction)
            return min_(phased_in, limitation)

    class ny_eitc(Variable):
        value_type = float
        entity = TaxUnit
        label = "New York EITC"
        unit = USD
        definition_period = YEAR
        reference = "https://legislation.nysenate.gov/pdf/bills/2023/S277C"
        defined_for = StateCode.NY

        def formula(tax_unit, period, parameters):
            # The EITC match percentage is phased out for amounts attributed to
            # younger children
            federal_eitc = tax_unit("younger_eitc", period)
            p = parameters(period).gov.contrib.states.ny.wftc.eitc
            tentative_nys_eitc = federal_eitc * p.match
            # The EITC match amount remains flat attributed to older children
            older_child_eitc = tax_unit("older_eitc", period)
            p_ny = parameters(period).gov.states.ny.tax.income.credits
            older_ny_eitc = older_child_eitc * p_ny.eitc.match
            total_credit = tentative_nys_eitc + older_ny_eitc
            household_credit = tax_unit("ny_household_credit", period)
            return max_(0, total_credit - household_credit)

    class ny_exemptions_dependent(Variable):
        value_type = bool
        entity = Person
        label = "Dependent under the New York exemptions definition"
        unit = USD
        definition_period = YEAR
        reference = "https://www.nysenate.gov/legislation/laws/TAX/616"
        defined_for = StateCode.NY

        def formula(person, period, parameters):
            p = parameters(period).gov.contrib.states.ny.wftc.exemptions
            if p.in_effect:
                child_dependent = person("is_child_dependent", period)
                wftc_eligible_child = person("ny_wftc_eligible_child", period)
                return child_dependent & ~wftc_eligible_child
            return person("is_child_dependent", period)

    class ny_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "NY exemptions"
        unit = USD
        definition_period = YEAR
        reference = "https://www.nysenate.gov/legislation/laws/TAX/616"
        defined_for = StateCode.NY

        def formula(tax_unit, period, parameters):

            count_dependents = add(
                tax_unit, period, ["ny_exemptions_dependent"]
            )
            dependent_exemption = parameters(
                period
            ).gov.states.ny.tax.income.exemptions.dependent
            return dependent_exemption * count_dependents

    class ny_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        label = "NY refundable credits"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.NY

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.ny.tax.income.credits
            older_credits = add(tax_unit, period, p.refundable)
            ny_wftc = tax_unit("ny_working_families_tax_credit", period)
            return older_credits + ny_wftc

    class reform(Reform):
        def apply(self):
            self.update_variable(ny_working_families_tax_credit)
            self.update_variable(ny_wftc_eligible_child)
            self.update_variable(ny_wftc_eligible_children)
            self.update_variable(is_younger_child_dependent)
            self.update_variable(eitc_younger_children_count)
            self.update_variable(eitc_younger_maximum)
            self.update_variable(eitc_younger_phase_in_rate)
            self.update_variable(eitc_younger_phased_in)
            self.update_variable(eitc_younger_phase_out_start)
            self.update_variable(eitc_younger_phase_out_rate)
            self.update_variable(eitc_younger_reduction)
            self.update_variable(eitc_younger_demographic_eligible)
            self.update_variable(younger_eitc_eligible)
            self.update_variable(younger_eitc)
            self.update_variable(is_older_child_dependent)
            self.update_variable(eitc_older_children_count)
            self.update_variable(eitc_older_maximum)
            self.update_variable(eitc_older_phase_in_rate)
            self.update_variable(eitc_older_phased_in)
            self.update_variable(eitc_older_phase_out_start)
            self.update_variable(eitc_older_phase_out_rate)
            self.update_variable(eitc_older_reduction)
            self.update_variable(eitc_older_demographic_eligible)
            self.update_variable(older_eitc_eligible)
            self.update_variable(older_eitc)
            self.update_variable(ny_eitc)
            self.neutralize_variable("ny_ctc")
            self.update_variable(ny_exemptions_dependent)
            self.update_variable(ny_exemptions)
            self.update_variable(ny_refundable_credits)

    return reform


def create_ny_working_families_tax_credit_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_ny_working_families_tax_credit()

    p = parameters(period).gov.contrib.states.ny.wftc

    if p.in_effect:
        return create_ny_working_families_tax_credit()
    else:
        return None


ny_working_families_tax_credit = create_ny_working_families_tax_credit_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class nj_dependents_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey qualified and other dependent children exemption"
    reference = "https://casetext.com/statute/new-jersey-statutes/title-54a-new-jersey-gross-income-tax-act/chapter-54a3-personal-exemptions-and-deductions/section-54a3-1-personal-exemptions-and-deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        # Then get the NJ Exemptions part of the parameter tree.
        p = parameters(period).gov.states.nj.tax.income.exemptions.dependents

        # Total the number of dependents.
        dependents = tax_unit("tax_unit_dependents", period)

        # Get their dependent exemption amount based on their filing status.
        return dependents * p.amount



from policyengine_us.model_api import *


class ny_supplemental_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY Supplemental EITC"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (d)(8)

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.ny.tax.income.credits
        return federal_eitc * p.eitc.supplemental_match



from policyengine_us.model_api import *


class heat_expense_included_in_rent(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Whether heating expense is included in rent payments"
    definition_period = YEAR
    # Excludes other types of utilities
    reference = "https://liheapch.acf.hhs.gov/tables/FY2016/subsidize.htm"



from policyengine_us.model_api import *


class dc_ccsp(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "DC Child Care Subsidy Program (CCSP) benefit amount"
    definition_period = MONTH
    reference = "https://osse.dc.gov/subsidy"
    defined_for = "dc_ccsp_eligible"

    def formula(spm_unit, period, parameters):
        copay = spm_unit("dc_ccsp_copay", period)
        maximum_payment = add(
            spm_unit, period, ["dc_ccsp_maximum_subsidy_amount"]
        )
        pre_subsidy_childcare_expense = spm_unit(
            "spm_unit_pre_subsidy_childcare_expenses", period
        )
        uncapped_subsidy_amount = max_(
            pre_subsidy_childcare_expense - copay, 0
        )
        return min_(uncapped_subsidy_amount, maximum_payment)



from policyengine_us.model_api import *


class ma_tafdc_earned_income_after_deductions(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) earned income after deductions"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-280"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        # The 50% disregard only applies the unit has received TAFDC in the past 4 months and
        # is re-applying due to temporary ineligibility
        gross_earned_income = add(
            spm_unit, period, ["ma_tcap_gross_earned_income"]
        )
        deductions = add(
            spm_unit,
            period,
            [
                "ma_tafdc_work_related_expense_deduction",
                "ma_tafdc_dependent_care_deduction",
            ],
        )
        return max_(gross_earned_income - deductions, 0)



from policyengine_us.model_api import *


class nj_pension_retirement_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey Pension/Retirement Exclusion"
    unit = USD
    documentation = "New Jersey pension and retirement excludable amount if eligible (Line 28a)"
    definition_period = YEAR
    reference = (
        "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=21",
        "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-6-10/",
    )
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        pension_income = add(tax_unit, period, ["nj_eligible_pension_income"])
        fraction = tax_unit("nj_retirement_exclusion_fraction", period)
        p = parameters(period).gov.states.nj.tax.income.exclusions.retirement
        filing_status = tax_unit("filing_status", period)
        return min_(pension_income * fraction, p.max_amount[filing_status])



from policyengine_us.model_api import *


class mn_child_and_working_families_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota child and working families credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revisor.mn.gov/statutes/cite/290.0661#stat.290.0661.4"
        "https://www.revisor.mn.gov/statutes/cite/290.0671"
        "https://www.revenue.state.mn.us/sites/default/files/2024-01/m1cwfc-23_1.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        # Minnesota combines the Child Tax Credit and
        # the Working Family Credit into a single credit
        # The credits are defined by separate legal code sections
        # but are calculated together in the tax forms.
        # Child Tax Credit computation:
        qualifying_children = tax_unit("ctc_qualifying_children", period)
        p = parameters(period).gov.states.mn.tax.income.credits.cwfc
        base_ctc_amount = qualifying_children * p.ctc.amount
        # Working Family Credit computation:
        wfc_eligible = tax_unit("mn_wfc_eligible", period)
        # The credit is phased in based on earnings
        earnings = tax_unit("filer_adjusted_earnings", period)
        base_wfc_credit = p.wfc.phase_in.calc(earnings)
        person = tax_unit.members
        qualifying_child = person("is_child_dependent", period)
        age = person("age", period)
        full_time_student = person("is_full_time_student", period)
        qualifying_older_child = (
            qualifying_child
            & (age > p.wfc.additional.age_threshold)
            & full_time_student
        )
        qualifying_older_children = tax_unit.sum(qualifying_older_child)
        additional_wfc_credit = p.wfc.additional.amount.calc(
            qualifying_older_children
        )
        base_wfc_amount = (
            base_wfc_credit + additional_wfc_credit
        ) * wfc_eligible
        # The credit amounts are combined and phase out together based on the number of
        # qualifying older children
        combined_credit = base_ctc_amount + base_wfc_amount
        # The phase out threshold is based on the larger of AGI and earnings and filing status
        agi = tax_unit("adjusted_gross_income", period)
        income = max_(earnings, agi)
        filing_status = tax_unit("filing_status", period)
        phase_out_threshold = where(
            filing_status == filing_status.possible_values.JOINT,
            p.phase_out.threshold.joint,
            p.phase_out.threshold.other,
        )
        excess_income = max_(0, income - phase_out_threshold)
        # Minnesota applies a different phase out rate for filers ineligible for the Child Tax Credit
        # with qualifying older children for the WFC
        lower_phase_out_rate_eligible = (qualifying_older_children > 0) & (
            base_ctc_amount == 0
        )
        phase_out_rate = where(
            lower_phase_out_rate_eligible,
            p.phase_out.rate.ctc_ineligible_with_qualifying_older_children,
            p.phase_out.rate.main,
        )
        reduction = excess_income * phase_out_rate
        return max_(0, combined_credit - reduction)



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_mt_ctc() -> Reform:
    class mt_ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Montana Child Tax Credit"
        definition_period = YEAR
        unit = USD
        reference = "https://leg.mt.gov/bills/2023/billpdf/HB0268.pdf"
        defined_for = "mt_ctc_eligible"

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.mt.ctc
            person = tax_unit.members
            age = person("age", period)
            credit_amount = tax_unit.sum(p.amount.calc(age))
            # Credit gets reduced by an amount for each increment that AGI exceeds a certain threshold
            agi = tax_unit("adjusted_gross_income", period)
            excess = max_(agi - p.reduction.threshold, 0)
            increments = excess // p.reduction.increment
            reduction = p.reduction.amount * increments
            return max_(credit_amount - reduction, 0)

    class mt_ctc_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Eligible for the Montana Child Tax Credit"
        definition_period = YEAR
        reference = "https://leg.mt.gov/bills/2023/billpdf/HB0268.pdf"
        defined_for = StateCode.MT

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.mt.ctc.income_limit
            agi = tax_unit("adjusted_gross_income", period)
            agi_eligible = agi <= p.agi
            # CTC limited to filers with investment income below a certain threshold
            investment_income_eligible = (
                tax_unit("eitc_relevant_investment_income", period)
                < p.investment
            )

            earned_income = tax_unit("tax_unit_earned_income", period)
            receives_earned_income = earned_income > 0

            return (
                agi_eligible
                & investment_income_eligible
                & receives_earned_income
            )

    def modify_parameters(parameters):
        parameters.gov.states.mt.tax.income.credits.refundable.update(
            start=instant("2023-01-01"),
            stop=instant("2031-12-31"),
            value=["mt_ctc", "mt_eitc"],
        )
        return parameters

    class reform(Reform):
        def apply(self):
            self.update_variable(mt_ctc)
            self.update_variable(mt_ctc_eligible)
            self.modify_parameters(modify_parameters)

    return reform


def create_mt_ctc_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_mt_ctc()

    p = parameters.gov.contrib.states.mt.ctc

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_mt_ctc()
    else:
        return None


mt_ctc = create_mt_ctc_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class nj_tanf_maximum_benefit(Variable):
    value_type = float
    entity = SPMUnit
    label = "New Jersey TANF maximum benefit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NJ

    def formula(spm_unit, period, parameters):
        # Get number of people in SPM unit.
        people = spm_unit("spm_unit_size", period)
        # Cap them at the maximum specified in the tables.
        capped_people = min_(people, 8).astype(int)
        # Calculate additional people beyond the maximum in tables.
        additional_people = people - capped_people
        # Get the relevant part of the parameter tree.
        p = parameters(period).gov.states.nj.njdhs.tanf.maximum_benefit
        # Look up the main maximum benefit for the number of (capped) people.
        base = p.main[capped_people]
        # Add the additional maximum benefit for the additional people.
        additional_maximum_benefit = p.additional * additional_people
        monthly = base + additional_maximum_benefit
        # Return annual value.
        return monthly * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class id_aged_or_disabled_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho aged or disabled deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        # The program includes individuals in the household, including those not in the filing unit.
        # We simplify by limiting to those in the filing unit.
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.id.tax.income.subtractions.aged_or_disabled
        eligible_person = person(
            "id_aged_or_disabled_deduction_eligible_person", period
        )
        total_eligible_people = tax_unit.sum(eligible_person)
        capped_eligible_people = min_(total_eligible_people, p.person_cap)
        deduction_amount = capped_eligible_people * p.amount
        # Filers cannot claim the subtraction if they claim the credit.
        eligible = ~tax_unit("id_receives_aged_or_disabled_credit", period)
        return eligible * deduction_amount



from policyengine_us.model_api import *


class vt_retirement_income_exemption_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Vermont retirement income exemption eligibility status"
    reference = (
        "https://legislature.vermont.gov/statutes/section/32/151/05811",  # Titl. 32 V.S.A. § 5811(21)(B)(iv)
        "https://legislature.vermont.gov/statutes/section/32/151/05830e"  # Titl. 32 V.S.A. § 5830e
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112%20Instr-2022.pdf#page=3",  # Instruction for 2022 SCHEDULE IN-112 - RETIREMENT INCOME EXEMPTION WORKSHEET
        "https://tax.vermont.gov/individuals/seniors-and-retirees",  # Instruction for exemption from different retirement system
    )
    defined_for = StateCode.VT
    documentation = "Vermont filers use below criteria to check whether the tax unit is eligible for vermont retirement income exemption."

    def formula(tax_unit, period, parameters):
        # Filer can choose from one of Social Security,
        # Civil Service Retirement System (CSRS), Military Retirement Income
        # or other eligible retirement systems to determine eligibility
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        p = parameters(
            period
        ).gov.states.vt.tax.income.agi.retirement_income_exemption.csrs.reduction
        # One of the retirement income should be greater than 0
        retirement_income = add(
            tax_unit,
            period,
            [
                "tax_unit_taxable_social_security",
                "military_retirement_pay",
                "csrs_retirement_pay",
            ],
        )
        retirement_income_qualified = retirement_income > 0
        # The agi should below threshold
        agi_qualified = agi < p.end[filing_status]
        # Both qualified then the filer is qualified for vermont retirement
        # income exemption
        return retirement_income_qualified & agi_qualified



from policyengine_us.model_api import *


class hi_interest_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii interest deduction"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=17"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=32"  # total itemized deduction worksheet
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    adds = ["investment_interest_expense", "mortgage_interest"]



from policyengine_us.model_api import *


class is_surviving_spouse(Variable):
    value_type = bool
    entity = Person
    label = "surviving spouse"
    documentation = "Whether the person is surviving spouse."
    definition_period = YEAR



from policyengine_us.model_api import *


class ms_income_tax_before_credits_unit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi income tax before credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MS

    def formula(tax_unit, period, parameters):
        ms_files_separately = tax_unit("ms_files_separately", period)
        itax_indiv = add(
            tax_unit, period, ["ms_income_tax_before_credits_indiv"]
        )
        itax_joint = add(
            tax_unit, period, ["ms_income_tax_before_credits_joint"]
        )
        return where(ms_files_separately, itax_indiv, itax_joint)



from policyengine_us.model_api import *


class tax_exempt_403b_distributions(Variable):
    value_type = float
    entity = Person
    label = "tax-exempt 403(b) distributions"
    unit = USD
    documentation = (
        "Tax-exempt distributions from 403(b) accounts (typically Roth)."
    )
    definition_period = YEAR



from policyengine_us.model_api import *


class id_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho taxable income"
    unit = USD
    documentation = "Idaho taxable income"
    definition_period = YEAR
    reference = "https://tax.idaho.gov/wp-content/uploads/forms/EFO00089/EFO00089_12-30-2022.pdf#page=1"
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        agi = tax_unit("id_agi", period)
        deductions = tax_unit("id_deductions", period)
        qbid = tax_unit("qualified_business_income_deduction", period)
        return max_(0, agi - deductions - qbid)



from policyengine_us.model_api import *


class ma_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA EITC"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.mass.gov/info-details/mass-general-laws-c62-ss-6"  # (h)
    )
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        rate = parameters(period).gov.states.ma.tax.income.credits.eitc.match
        return federal_eitc * rate



from policyengine_us.model_api import *


class or_kicker(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR Kicker"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=19",
        "https://www.oregonlegislature.gov/bills_laws/Pages/OrConst.aspx",  # Article IX Section 14 (4)
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["or"].tax.income.credits.kicker
        return p.percent * tax_unit(
            "or_tax_before_credits_in_prior_year", period
        )



from policyengine_us.model_api import *


class used_clean_vehicle_credit_credit_limit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Used clean vehicle credit credit limit"
    documentation = "Nonrefundable credit for the purchase of a previously-owned clean vehicle"
    unit = USD
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=370"
    defined_for = "used_clean_vehicle_credit_eligible"

    def formula(tax_unit, period, parameters):
        income_tax_before_credits = tax_unit(
            "income_tax_before_credits", period
        )
        p = parameters(period).gov.irs.credits.clean_vehicle.used
        preceding_credits = add(tax_unit, period, p.preceding_credits)
        return max_(income_tax_before_credits - preceding_credits, 0)



from policyengine_us.model_api import *


class co_oap_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Colorado Old Age Pension Eligible"
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        assets = person("ssi_countable_resources", period)
        joint_claim = person("ssi_claim_is_joint", period)
        p = parameters(period).gov.states.co.ssa.oap
        asset_limit = where(
            joint_claim, p.resources.couple, p.resources.single
        )
        below_asset_limit = assets <= asset_limit
        age = person("age", period)
        in_age_range = p.age_range.calc(age)
        return below_asset_limit & in_age_range



from policyengine_us.model_api import *


class spm_unit_is_in_spm_poverty(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SPM unit in SPM poverty"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        income = spm_unit("spm_unit_net_income", period)
        poverty_threshold = spm_unit("spm_unit_spm_threshold", period)
        return income < poverty_threshold



from policyengine_us.model_api import *


class mo_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Sum of itemized deductions applicable to Missouri taxable income calculation"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf",
        "https://dor.mo.gov/forms/4711_2021.pdf#page=11",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.141&bid=7212",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mo.tax.income
        deduction_value = add(tax_unit, period, p.deductions.itemized)
        net_state_income_taxes = tax_unit("mo_net_state_income_taxes", period)
        return max_(0, deduction_value - net_state_income_taxes)



from policyengine_us.model_api import *


class aged_spouse(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Aged spouse"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/63#f"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.standard.aged_or_blind
        return tax_unit("age_spouse", period) >= p.age_threshold



from policyengine_us.model_api import *


class eitc_demographic_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Meets demographic eligibility for EITC"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/32#c_1_A"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        has_child = tax_unit("eitc_child_count", period) > 0
        age = person("age", period)
        # Relative parameter reference break branching in some states that
        # modify EITC age limits.
        min_age_non_student = (
            parameters.gov.irs.credits.eitc.eligibility.age.min(period)
        )
        min_age_student = (
            parameters.gov.irs.credits.eitc.eligibility.age.min_student(period)
        )
        student = person("is_full_time_student", period)
        min_age = where(student, min_age_student, min_age_non_student)
        max_age = parameters.gov.irs.credits.eitc.eligibility.age.max(period)
        meets_age_requirements = (age >= min_age) & (age <= max_age)
        return has_child | tax_unit.any(meets_age_requirements)



from policyengine_us.model_api import *


class nm_cdcc_max_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico maximum credit for dependent child day care credit"
    unit = USD
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503752/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgDsfAEwBKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = "nm_cdcc_eligible"

    def formula(tax_unit, period, parameters):
        # allocate unit's childcare expenses across eligible children
        person = tax_unit.members
        expenses = person.tax_unit("tax_unit_childcare_expenses", period)
        eligible = person("nm_cdcc_eligible_child", period).astype(float)
        count_eligible = person.tax_unit.sum(eligible)
        expense = np.zeros_like(eligible)
        mask = count_eligible > 0
        expense[mask] = eligible[mask] * expenses[mask] / count_eligible[mask]
        # cap each eligible child's allowable expense
        p = parameters(period).gov.states.nm.tax.income.credits.cdcc
        child_expense = min_(p.max_amount.per_child, p.rate * expense)
        # cap unit's total allowable expense
        total_expense = tax_unit.sum(child_expense)
        return min_(p.max_amount.total, total_expense)



from policyengine_us.model_api import *


class ky_tuition_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kentucky tuition tax credit"  # Form 8863-K
    unit = USD
    definition_period = YEAR
    defined_for = "ky_tuition_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        # We do not currently model the limitation to undergraduate studies at Kentucky institutions.
        tentative_tax_credit = add(
            tax_unit,
            period,
            ["american_opportunity_credit", "lifetime_learning_credit"],
        )
        rate = parameters(
            period
        ).gov.states.ky.tax.income.credits.tuition_tax.rate
        return tentative_tax_credit * rate



from policyengine_us.model_api import *


class military_service_income(Variable):
    value_type = float
    entity = Person
    label = "Military service income"
    unit = USD
    documentation = "Military pay from active duty, National Guard, and/or the reserve component of the armed forces."
    definition_period = YEAR



from policyengine_us.model_api import *


class dc_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=18"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=17"
    )
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        # follows Calculation F in references
        # calculate US itemized deductions less state non-property taxes
        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        uncapped_property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        # calculate DC itemized deductions before partial phase-out
        dc_itm_deds = itm_deds_less_salt + uncapped_property_taxes
        # apply partial phase-out of DC itemized deductions
        EXEMPT_ITEMS = [
            "medical_expense_deduction",
            "casualty_loss_deduction",
        ]
        exempt_deds = add(tax_unit, period, EXEMPT_ITEMS)
        nonexempt_deds = max_(0, dc_itm_deds - exempt_deds)
        dc_agi = add(tax_unit, period, ["dc_agi"])
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.dc.tax.income.deductions
        phase_out_start = p.itemized.phase_out.start[filing_status]
        excess_agi = max_(0, dc_agi - phase_out_start)
        phase_out_amount = excess_agi * p.itemized.phase_out.rate
        limited_nonexempt_deds = max_(0, nonexempt_deds - phase_out_amount)
        return exempt_deds + limited_nonexempt_deds



from policyengine_us.model_api import *


class ca_tanf_region1(Variable):
    value_type = bool
    entity = Household
    label = "In a CalWORKs region 1 county"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FCalWORKs%2FCalWORKs%2F44-212_Minimum_Basic_Standard_of_Adequate_Care%2F44-212_Minimum_Basic_Standard_of_Adequate_Care.htm%23Definitionsbc-4&rhtocid=_3_1_7_20_3"

    def formula(household, period, parameters):
        county = household("county_str", period)
        region1_counties = parameters(
            period
        ).gov.states.ca.cdss.tanf.cash.region1_counties
        return np.isin(county, region1_counties)



from policyengine_us.model_api import *


class hud_very_low_income_factor(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD Very Low income factor"
    unit = USD
    documentation = "Very Low income factor for HUD programs"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.hud.ami_limit
        size = spm_unit("spm_unit_size", period)
        p.family_size_excess = p.per_person_exceeding_4
        size_exceeding_4 = max_(size - 4, 0)
        size_capped_at_4 = min_(size, 4)
        return (
            p.family_size.VERY_LOW[size_capped_at_4]
            + p.family_size_excess.VERY_LOW * size_exceeding_4
        )



from policyengine_us.model_api import *


class ri_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island taxable income"
    reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf"
    defined_for = StateCode.RI
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        ri_exemptions = tax_unit("ri_exemptions", period)
        # Modified Federal AGI
        mod_agi = tax_unit("ri_agi", period)
        ri_deductions = tax_unit("ri_standard_deduction", period)
        agi_less_deductions = max_(mod_agi - ri_deductions, 0)
        return max_(agi_less_deductions - ri_exemptions, 0)



from policyengine_us.model_api import *


class electricity_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "Electricity expense"
    unit = USD
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        pre_subsidy_electricity_expenses = spm_unit(
            "pre_subsidy_electricity_expense", period
        )
        p = parameters(period).household.expense.utilities
        subsidies = add(spm_unit, period, p.subsidies)
        return max_(pre_subsidy_electricity_expenses - subsidies, 0)



from policyengine_us.model_api import *


class CHIPCategory(Enum):
    CHILD = "Child"
    PREGNANT_STANDARD = "Pregnant - Standard"
    PREGNANT_FCEP = "Pregnant - FCEP"
    NONE = "None"


class chip_category(Variable):
    value_type = Enum
    possible_values = CHIPCategory
    default_value = CHIPCategory.NONE
    entity = Person
    label = "CHIP category"
    documentation = "Category under which a person is eligible for the Children's Health Insurance Program"
    definition_period = YEAR
    reference = (
        "https://www.ssa.gov/OP_Home/ssact/title21/2110.htm",
        "https://www.medicaid.gov/medicaid/national-medicaid-chip-program-information/medicaid-childrens-health-insurance-program-basic-health-program-eligibility-levels",
        "https://www.kff.org/affordable-care-act/state-indicator/medicaid-and-chip-income-eligibility-limits-for-pregnant-women-as-a-percent-of-the-federal-poverty-level",
    )

    def formula(person, period, parameters):
        is_child_eligible = person("is_chip_eligible_child", period)
        is_chip_eligible_standard_pregnant_person = person(
            "is_chip_eligible_standard_pregnant_person", period
        )
        is_chip_fcep_eligible_person = person(
            "is_chip_fcep_eligible_person", period
        )

        # Use select to return the appropriate category
        # If eligible under multiple categories, prioritize child, then standard pregnant, then FCEP
        return select(
            [
                is_child_eligible,
                is_chip_eligible_standard_pregnant_person,
                is_chip_fcep_eligible_person,
            ],
            [
                CHIPCategory.CHILD,
                CHIPCategory.PREGNANT_STANDARD,
                CHIPCategory.PREGNANT_FCEP,
            ],
            default=CHIPCategory.NONE,
        )



from policyengine_us.model_api import *


class la_general_relief_age_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    label = "Eligible for the Los Angeles County General Relief based on the age requirements"
    # Person has to be a resident of LA County
    defined_for = "in_la"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.ca.la.general_relief.eligibility
        # Based on the fact sheet and questionnaire, assuming that just the head
        # has to be over the age threshold
        age = add(spm_unit, period, ["age_head"])
        # Person has to reach a certain age
        return age >= p.age_threshold



from policyengine_us.model_api import *


class is_ca_cvrp_normal_rebate_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Eligible for CVRP normal rebate"
    documentation = "Eligible for California Clean Vehicle Rebate Project (CVRP) normal rebate"
    reference = "https://cleanvehiclerebate.org/en/eligibility-guidelines"
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        # AGI must be less than the threshold.
        agi = person.tax_unit("adjusted_gross_income", period)
        filing_status = person.tax_unit("filing_status", period)
        caps = parameters(period).gov.states.ca.calepa.carb.cvrp.income_cap
        return agi <= caps[filing_status]



from policyengine_us.model_api import *


class household_tax_before_refundable_credits(Variable):
    value_type = float
    entity = Household
    label = "total tax before refundable credits"
    documentation = "Total tax liability before refundable credits."
    unit = USD
    definition_period = YEAR
    adds = [
        "employee_payroll_tax",
        "self_employment_tax",
        "income_tax_before_refundable_credits",
        "flat_tax",
        "household_state_tax_before_refundable_credits",
    ]



from policyengine_us.model_api import *


class dc_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC child and dependent care credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=36"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=34"
    )
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        # DC matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        p_dc = parameters(period).gov.states.dc.tax.income.credits
        return us_cdcc * p_dc.cdcc.match



from policyengine_us.model_api import *


class tip_income_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tip income deduction"
    unit = USD
    definition_period = YEAR
    defined_for = "tip_income_deduction_ssn_requirement_met"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        tip_income = person("tip_income", period)
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.irs.deductions.tip_income
        start = p.phase_out.start[filing_status]
        agi_excess = max_(agi - start, 0)
        phase_out_amount = agi_excess * p.phase_out.rate
        total_tip_income = tax_unit.sum(tip_income)
        capped_tip_income = min_(p.cap, total_tip_income)
        return max_(0, capped_tip_income - phase_out_amount)



from policyengine_us.model_api import *


class id_capital_gains_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho capital gains deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://legislature.idaho.gov/statutesrules/idstat/title63/t63ch30/sect63-3022h/"
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.id.tax.income.deductions.capital_gains
        # taxpayer must report capital gain net income
        net_capital_gain = tax_unit("property_sales_net_capital_gain", period)

        # Ordinary Income do not qualify for the Idaho capital gains deduction
        # Gain from dispositions of certain depreciable property treated as ordinary income

        # Return calculated amount.
        return p.percentage * net_capital_gain



from policyengine_us.model_api import *


class hud_adjusted_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD adjusted income"
    unit = USD
    documentation = "Adjusted income for HUD programs"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/24/5.611"

    def formula(spm_unit, period, parameters):
        # Extract annual income.
        income = spm_unit("hud_annual_income", period)
        # Count dependents - children only for now.
        child_count = add(spm_unit, period, ["is_child"])
        # Identify if elderly or disabled.
        elderly_disabled = spm_unit("is_hud_elderly_disabled_family", period)
        # Extract childcare expenses.
        # "Any reasonable child care expenses necessary to enable a member of
        # the family to be employed or to further his or her education."
        childcare_expenses = spm_unit("childcare_expenses", period)
        # TODO: Attendant care (save for later)
        # Medical expenses for elderly/disabled families.
        ded = parameters(period).gov.hud.adjusted_income.deductions
        moop = add(spm_unit, period, ["medical_out_of_pocket_expenses"])
        # Only expenses beyond a percent of income are deductible.
        moop_threshold = ded.moop.threshold * income
        moop_deductible = max_(0, moop - moop_threshold)
        moop_ded = moop_deductible * elderly_disabled
        # Add dependent and elderly disabled deductions.
        dependent_ded = ded.dependent.amount * child_count
        elderly_disabled_ded = ded.elderly_disabled.amount * elderly_disabled
        # Calculate and return adjusted income, non-negative.
        return max_(
            income
            - dependent_ded
            - elderly_disabled_ded
            - childcare_expenses
            - moop_ded,
            0,
        )



from policyengine_us.model_api import *


class is_ccdf_asset_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "Asset eligibility for CCDF"

    def formula(spm_unit, period, parameters):
        assets = spm_unit("spm_unit_assets", period)
        p_asset_limit = parameters(period).gov.hhs.ccdf.asset_limit
        return assets <= p_asset_limit
